"""
A simple repeat with two alternative endings (volta brackets).
"""

from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase


class TestLily45b(IdTestCase):
    def test_lily_45b_RepeatWithAlternatives(self):
        score = Score()
        part = score.add_part('p1')
        [part.add_chord(Chord(72, 4)) for _ in range(4)]
        part.get_measure(2).set_repeat_barline()
        part.get_measure(2).set_repeat_ending(number=1, type='start')
        part.get_measure(2).set_repeat_ending(number=1, type='stop')
        part.get_measure(3).set_repeat_ending(number=2, type='start')
        part.get_measure(3).set_repeat_ending(number=2, type='discontinue')
        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloWorld(IdTestCase):
    def test_export_hello_world(self):
        """
        Hello World as musicxml means having a C4 pitch as a whole in a 4/4 measure with treble clef.
        """
        """
        Tester creates score
        """
        s = Score()
        """
        He adds a part with id P1 (required) and name Music
        """
        p = s.add_child(Part('P1', name='Music'))
        """He adds a Chord with midi 60 to the part. A measure (default Measure has a 4/4 time signature) will be 
        created automatically"""
        p.add_chord(Chord(60, 4))
        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.quarterduration import QuarterDuration
from musicscore.score import Score
from musicscore.tests.util import IdTestCase
from musicscore.util import lcm


class TestQuantization(IdTestCase):
    def test_get_and_set_possible_subdivisions(self):
        s = Score()
        p = s.add_child(Part('p1'))
        m = p.add_measure()
        s = m.get_children()[0]
        v = s.get_children()[0]
        beats = v.get_children()
        for beat in beats:
            assert beat.get_possible_subdivisions() == [2, 3, 4, 5, 6, 7, 8]
        beats[0].quarter_duration = 1 / 2
        assert beats[0].get_possible_subdivisions() == [2, 3, 4, 5]
        beats[0].quarter_duration = 1 / 4
        assert beats[0].get_possible_subdivisions() == [2, 3]
        beats[0].quarter_duration = 1 / 8
        assert beats[0].get_possible_subdivisions() is None
        v.set_possible_subdivisions([2, 3, 4, 6], beat_quarter_duration=1)
        beats[1].set_possible_subdivisions([2, 4])
        assert beats[0].get_possible_subdivisions() is None
        assert beats[1].get_possible_subdivisions() == [2, 4]
        for beat in beats[2:]:
            assert beat.get_possible_subdivisions() == [2, 3, 4, 6]

        with self.assertRaises(ValueError):
            beats[1].set_possible_subdivisions([2, 4], beat_quarter_duration=0.5)

    def test_quantize_attribute(self):
        s = Score()
        p = s.add_child(Part('p1'))
        m = p.add_measure()

        st = m.get_children()[0]

        v = st.get_children()[0]
        beats = v.get_children()

        assert s.get_quantized is False
        assert p.get_quantized is False
        assert m.get_quantized is False
        assert st.get_quantized is False
        assert v.get_quantized is False
        for b in beats:
            assert b.get_quantized is False

        s.get_quantized = True
        assert p.get_quantized is True
        assert m.get_quantized is True
        assert st.get_quantized is True
        assert v.get_quantized is True
        for b in beats:
            assert b.get_quantized is True

        st.get_quantized = False
        assert s.get_quantized is True
        assert p.get_quantized is True
        assert m.get_quantized is True
        assert st.get_quantized is False
        assert v.get_quantized is False
        for b in beats:
            assert b.get_quantized is False

        p.get_quantized = False
        m.get_quantized = True
        st.get_quantized = False
        v.get_quantized = True
        beats[0].get_quantized = False
        beats[1].get_quantized = True
        beats[2].get_quantized = False
        beats[3].get_quantized = True

        assert s.get_quantized is True
        assert p.get_quantized is False
        assert m.get_quantized is True
        assert st.get_quantized is False
        assert v.get_quantized is True

    def test_simple_quantization(self):
        s = Score()
        p = s.add_child(Part('p1'))
        p.add_measure()
        p.add_chord(Chord(60, 0.2))
        p.add_chord(Chord(60, 3.8))
        assert [ch.quarter_duration for ch in p.get_chords()] == [0.2, 0.8, 3]
        b = p.get_measure(1).get_staff(1).get_voice(1).get_beat(1)
        b.set_possible_subdivisions(subdivisions=[2, 3, 4])
        b.quantize_quarter_durations()
        assert [ch.quarter_duration for ch in p.get_measure(1).get_staff(1).get_voice(1).get_chords()] == [0.25, 0.75,
                                                                                                           3]
        assert [ch.offset for ch in p.get_measure(1).get_staff(1).get_voice(1).get_chords()] == [0, 0.25, 0]

    def test_complex_quantization(self):
        s = Score()
        s.set_possible_subdivisions([2, 3, 4, 6, 8])
        p = s.add_child(Part('p1'))
        import random
        random.seed(11)
        quarter_durations = []
        while sum(quarter_durations) < 40:
            quarter_durations.append(QuarterDuration(random.random() + random.randint(0, 3)))
        quarter_durations.append(44 - sum(quarter_durations))

        for qd in quarter_durations:
            p.add_chord(Chord(midis=60, quarter_duration=qd))
        for beat in [b for m in p.get_children() for st in m.get_children() for v in st.get_children() for b in
                     v.get_children()]:
            beat.quantize_quarter_durations()
        # p.get_quantized = True
        for measure in p.get_children():
            v = measure.get_staff(1).get_voice(1)
            for b in v.get_children():
                qds = [ch.quarter_duration for ch in b.get_children()]
                div = lcm([qd.denominator for qd in qds])
                if div != 1:
                    assert div in [2, 3, 4, 6, 8]

    def test_part_quantization(self):
        s = Score()
        s.set_possible_subdivisions([2, 3, 4, 6, 8])
        p = s.add_child(Part('p1'))
        for qd in [2, 2 / 7, 1, 5 / 7]:
            p.add_chord(Chord(60, qd))
        p.get_quantized = True
        p.finalize()
        assert [ch.quarter_duration for ch in p.get_chords()] == [2, 1 / 4, 3 / 4, 1 / 4, 3 / 4]
        assert p.get_measure(1).get_divisions() == 4
        assert [ch.notes[0].xml_duration.value_ for ch in p.get_chords()] == [8, 1, 3, 1, 3]



from typing import List, Union
from quicktions import Fraction
import numbers

__all__ = ['BEATWISE_EXCEPTIONS', 'QuarterDuration', 'QuarterDurationMixin']

from musicscore.exceptions import QuarterDurationIsNotWritable

#: {offset: {quarter_duration: return value(s), ... }, ...}
BEATWISE_EXCEPTIONS = {0: {5: (3, 2), 6: (6,)}}

_note_simple_types = {
    (1, 12): '32nd',
    (1, 11): '32nd',
    (2, 11): '16th',
    (4, 11): 'eighth',
    (8, 11): 'quarter',
    (1, 10): '32nd',
    (1, 9): '32nd',
    (2, 9): '16th',
    (4, 9): 'eighth',
    (8, 9): 'quarter',
    (1, 8): '32nd',
    (1, 7): '16th',
    (2, 7): 'eighth',
    (4, 7): 'quarter',
    (1, 6): '16th',
    (1, 5): '16th',
    (2, 5): 'eighth',
    (4, 5): 'quarter',
    (8, 5): 'half',
    (1, 4): '16th',
    (2, 4): 'eighth',
    (1, 3): 'eighth',
    (2, 3): 'quarter',
    (4, 3): 'half',
    (1, 2): 'eighth',
    (1, 1): 'quarter',
    (2, 1): 'half',
    (4, 1): 'whole',
    (8, 1): 'breve',
}


class QuarterDuration(numbers.Rational):
    """
    A Class specifically designed for durations measured in quarters. The core of this class is a value of type quicktions.Fraction with a
    denominator limit of 1000, thus it can manage conversion of floats to fractions without usual inaccuracies of quintuples etc. See
    value property for more information.
    QuarterDuration has all needed magic methods for numeral comparison and conversion.
    """

    def __init__(self, *value):
        self._value = None
        self.value = value

    @property
    def numerator(self):
        """
        :return: Fraction's numerator.
        :rtype: int

        >>> QuarterDuration(1, 6).numerator
        1
        """
        return self.value.numerator

    @property
    def denominator(self):
        """
        :return: Fraction's denominator.
        :rtype: int

        >>> QuarterDuration(1, 6).denominator
        6
        """
        return self.value.denominator

    @property
    def value(self):
        """
        :return: QuarterDuration's value
        :rtype: quicktions.Fraction() with limit_denominator(1000)

        >>> QuarterDuration(3, 7).value
        Fraction(3, 7)
        >>> QuarterDuration(0.2).value
        Fraction(1, 5)
        >>> QuarterDuration(Fraction(1, 5)).value
        Fraction(1, 5)
        >>> QuarterDuration(1/5).value
        Fraction(1, 5)
        """
        return self._value

    @value.setter
    def value(self, val):
        if val is None or val == ():
            self._value = None
        elif isinstance(val, Fraction):
            self._value = val
        elif isinstance(val, str):
            self._value = Fraction(val).limit_denominator(1000)
        elif hasattr(val, '__iter__'):
            if len(val) == 1:
                self._value = Fraction(val[0]).limit_denominator(1000)
            elif len(val) == 2:
                self._value = Fraction(*val).limit_denominator(1000)
            else:
                raise ValueError

    def _get_type_and_dots(self):
        type = _note_simple_types.get(self.as_integer_ratio())
        if type:
            return type, 0
        else:
            qd = QuarterDuration(self.value * 2 / 3)
            type = _note_simple_types.get(qd.as_integer_ratio())
            if type:
                return type, 1
            else:
                qd = QuarterDuration(self.value * 4 / 7)
                type = _note_simple_types.get(qd.as_integer_ratio())
                if type:
                    return type, 2
                else:
                    raise QuarterDurationIsNotWritable(f'quarter duration {self} is not writable.')

    def as_integer_ratio(self):
        """
        :return: (numerator, denominator)
        :rtype: tuple

        >>> QuarterDuration(1, 5).as_integer_ratio()
        (1, 5)
        """
        return self.value.as_integer_ratio()

    def get_beatwise_sections(self, beats: List['Beat'], offset: Union[int, float, 'QuarterDuration', 'Fraction'] = 0):
        """
        :param beats:
        :param offset: offset in the first beat
        :return: [sections as list of QuarterDurations, leftover as QuarterDruation] leftover is the remaining quarter_duration which
                 exceeds the sum of all beats quarter durations. If there is no left over the second value in the list is None.
                 offset=0.5, beats=[Beat(1), Beat(1), Beat(1), Beat(1)] => [[0.5, 3], None]
                 See tests for more examples. In BEATWISE_EXCEPTIONS exceptions can be declared.
        """

        def _check_for_exception():
            exception = BEATWISE_EXCEPTIONS.get(offset)
            if exception:
                out = exception.get(sum(output[0]))
                if out:
                    output[0] = [QuarterDuration(value) for value in out]
                return output

        output = [None, [], None]
        if offset:
            output[0] = beats[0].quarter_duration - offset
            beats.pop(0)
        leftover = self - output[0] if output[0] is not None else self
        out_of_reach = leftover - sum(b.quarter_duration for b in beats)
        if out_of_reach > 0:
            leftover -= out_of_reach
            output[2] = out_of_reach
        for beat in beats:
            if leftover >= beat.quarter_duration:
                current_value = beat.quarter_duration
                leftover -= current_value
            else:
                current_value = leftover
                leftover = 0
            if not output[1]:
                output[1] = [current_value]
            else:
                if _is_writable(output[1][-1] + current_value):
                    output[1][-1] += current_value
                else:
                    output[1].append(current_value)
            if leftover == 0:
                break
        if output[0]:
            output = [[output[0], *output[1]], output[2]]
        else:
            output = [output[1], output[2]]

        exception = _check_for_exception()
        if exception:
            return exception

        return output

    def get_type(self):
        return self._get_type_and_dots()[0]

    def get_number_of_dots(self):
        return self._get_type_and_dots()[1]

    def __repr__(self):
        return f'{self.value.numerator}/{self.value.denominator}'

    def __str__(self):
        return f'QuarterDuration: {str(self.value)}'

    def __abs__(self):
        return QuarterDuration(self.value.__abs__())

    def __add__(self, other):
        return QuarterDuration(self.value.__add__(_convert_other(other)))

    def __ceil__(self):
        return QuarterDuration(self.value.__ceil__())

    def __floor__(self):
        return QuarterDuration(self.value.__floor__())

    def __floordiv__(self, other):
        return QuarterDuration(self.value.__floordiv__(_convert_other(other)))

    def __gt__(self, other):
        return self.value.__gt__(_convert_other(other))

    def __ge__(self, other):
        return self.value.__ge__(_convert_other(other))

    def __hash__(self):
        return self.value.__hash__()

    def __le__(self, other):
        return self.value.__le__(_convert_other(other))

    def __lt__(self, other):
        return QuarterDuration(self.value.__lt__(_convert_other(other)))

    def __mod__(self, other):
        return QuarterDuration(self.value.__mod__(_convert_other(other)))

    def __mul__(self, other):
        return QuarterDuration(self.value.__mul__(_convert_other(other)))

    def __neg__(self):
        return self.value.__neg__()

    def __pos__(self):
        return self.value.__pos__()

    def __pow__(self, power, modulo=None):
        return QuarterDuration(self.value.__pos__(power, modulo))

    def __radd__(self, other):
        return QuarterDuration(self.value.__radd__(_convert_other(other)))

    def __rfloordiv__(self, other):
        return QuarterDuration(self.value.__rfloordiv__(_convert_other(other)))

    def __rmod__(self, other):
        return QuarterDuration(self.value.__rmod__(_convert_other(other)))

    def __rmul__(self, other):
        return QuarterDuration(self.value.__rmul__(_convert_other(other)))

    def __round__(self, n=None):
        return QuarterDuration(self.value.__round__(n))

    def __rpow__(self, other):
        return QuarterDuration(self.value.__rpow__(_convert_other(other)))

    def __rtruediv__(self, other):
        return QuarterDuration(self.value.__rtruediv__(_convert_other(other)))

    def __truediv__(self, other):
        return QuarterDuration(self.value.__truediv__(_convert_other(other)))

    def __trunc__(self):
        return self.value.__trunc__()

    def __eq__(self, other):
        return self.value.__eq__(_convert_other(other))

    def __copy__(self):
        return self.__class__(self.value)

    def __deepcopy__(self, memodict={}):
        return self.__class__(self.value)


def _is_writable(quarter_duration: Union[float, int, Fraction, 'QuarterDuration']):
    """
    Function to check if a quarter duration is writable or must be split into two durations.

    :param quarter_duration:
    :return: boolean

    >>> _is_writable(5)
    False
    >>> _is_writable(7/8)
    False
    >>> _is_writable(3/8)
    True
    """
    writables = {1 / 64,
                 1 / 32,
                 3 / 64,
                 1 / 16,
                 3 / 32,
                 1 / 8,
                 3 / 16,
                 1 / 4,
                 3 / 8,
                 1 / 2,
                 3 / 4,
                 1,
                 3 / 2,
                 2,
                 3,
                 4,
                 6,
                 8,
                 12}

    if quarter_duration in writables:
        return True
    else:
        return False


def _convert_other(other):
    if isinstance(other, QuarterDuration):
        return other.value

    return Fraction(other).limit_denominator(1000)


class QuarterDurationMixin:
    """
    Mixin for all Classes with a quarter_duration. Used in :obj:`~musicscore.note.Note`, :obj:`~musicscore.chord.Chord` and
    :obj:`~musicscore.beat.Beat`
    """

    def __init__(self, quarter_duration=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._quarter_duration = None
        self.quarter_duration = quarter_duration

    def _set_quarter_duration(self, val):
        if isinstance(val, QuarterDuration):
            self._quarter_duration = val
        else:
            _check_quarter_duration_value(val)
            self._quarter_duration = QuarterDuration(val)

    @property
    def quarter_duration(self) -> QuarterDuration:
        """
        Duration measured in quarters.

        val type: Optional[int, float, Fraction, QuarterDuration]
        """
        return self._quarter_duration

    @quarter_duration.setter
    def quarter_duration(self, val):
        if val is not None:
            self._set_quarter_duration(val)
        else:
            self._quarter_duration = None


def _check_quarter_duration_value(val):
    if not isinstance(val, int) and not isinstance(val, float) and not isinstance(val, Fraction) and not isinstance(val,
                                                                                                                    QuarterDuration):
        raise TypeError(f'Wrong type for quarter duration {val}: {type(val)}')

    if val < 0:
        raise ValueError()

    return True



from pathlib import Path

from musicscore import Score, C, D, E, F, Chord, Midi
from musicscore.tests.util import IdTestCase

"""
Some microtones: c flat-and-a-half, d half-flat, e half-sharp, f sharp-and-a half. Once in the
lower and once in the upper region of the staff.
"""


class TestLily01d(IdTestCase):
    def test_lily_01d_Pitches_Microtones(self):
        score = Score('Microtones')
        p = score.add_part('microtones')
        midis = [Midi(60 - 3 / 2), Midi(60 + 3 / 2), Midi(64 + 1 / 2), Midi(65 + 3 / 2)]
        midis[0].accidental.mode = 'force-flat'
        midis[1].accidental.mode = 'force-flat'
        midis += [C(5, 'three-quarters-flat'), D(5, 'quarter-flat'), E(5, 'quarter-sharp'),
                  F(5, 'three-quarters-sharp')]

        for midi in midis:
            p.add_chord(Chord(midi, 1))

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



import cProfile
import timeit
from contextlib import redirect_stdout
from pathlib import Path
from xml.etree import ElementPath

from musicscore import Score, Part, Chord, Id
from musicxml.xmlelement.xmlelement import XMLScorePartwise, XMLLeftMargin


def hello_xml_score():
    XMLScorePartwise()


def time_hello_xml_score():
    print(timeit.timeit(hello_xml_score, number=1000))


with open(str(Path(__file__)) + 'xml_score.txt', '+w') as f:
    with redirect_stdout(f):
        cProfile.run('time_hello_xml_score()', sort="tottime")



from unittest import TestCase
from unittest.mock import patch

from musicscore.chord import Chord
from musicscore.measure import Measure
from musicscore.part import Part
from musicscore.staff import Staff
from musicscore.voice import Voice
from musicxml import XMLVoice, XMLStaff


class TestStaff(TestCase):
    def test_staff_init(self):
        st = Staff()
        assert st.number is None
        assert st.xml_object.value_ == 1
        st = Staff(number=3)
        assert st.xml_object.value_ == 3
        assert st.number == 3
        st.number = 2
        assert st.xml_object.value_ == 2
        st.xml_object.value_ = 1
        assert st.number is 1
        assert st.xml_object.value_ == 1

    @patch('musicscore.measure.Measure')
    def test_add_child(self, mock_measure):
        st = Staff()
        st._parent = mock_measure
        assert [child.number for child in st.get_children()] == []
        st.add_child(Voice())
        assert [child.number for child in st.get_children()] == [1]
        st.add_child(Voice())
        assert len(st.get_children()) == 2
        assert [child.number for child in st.get_children()] == [1, 2]

    @patch('musicscore.measure.Measure')
    def test_add_voice(self, mock_measure):
        st = Staff()
        st._parent = mock_measure
        v = st.add_voice()
        assert st.get_children()[-1] == v
        assert len(st.get_children()) == 1
        assert v.number == 1
        v = st.add_voice()
        assert st.get_children()[-1] == v
        assert len(st.get_children()) == 2
        assert v.number == 2

        v = st.add_voice(5)
        assert v.number == 5
        assert len(st.get_children()) == 5
        assert [v.number for v in st.get_children()] == [1, 2, 3, 4, 5]

    def test_get_previous_staff(self):
        p = Part('P1')
        m1 = p.add_measure()
        m2 = p.add_measure()
        m1.add_staff(2)
        m2.add_staff(2)
        st11, st12 = m1.get_children()
        st21, st22 = m2.get_children()
        assert st11.get_previous_staff() is None
        assert st12.get_previous_staff() is None
        assert st21.get_previous_staff() == st11
        assert st22.get_previous_staff() == st12

    def test_get_last_steps_with_accidentals(self):
        m1 = Measure(1)
        m1._add_chord(Chord(midis=[61, 62, 63], quarter_duration=2))
        m1._add_chord(Chord(midis=[63, 64, 66], quarter_duration=2))
        assert m1.get_staff(1).get_last_pitch_steps_with_accidentals() == {'E', 'F'}

    def test_staff_clef(self):
        st = Staff()
        assert st.clef is None

    def test_get_staff_number_from_midi(self):
        part = Part('p1')
        ch = Chord([60, 70, 80], 2)
        part.add_chord(ch)
        ch.midis[0].set_staff_number(2)
        ch.midis[1].set_staff_number(2)
        part.finalize()

        for index, n in enumerate(ch.notes):
            if index in [0, 1]:
                assert n.xml_object.xml_staff.value_ == 2
            else:
                assert not n.xml_object.xml_staff



import copy
from unittest import skip, TestCase

from quicktions import Fraction

from musicscore import BassClef, Score, Part
from musicscore.accidental import Accidental
from musicscore.beat import Beat
from musicscore.chord import Chord, _split_copy, _group_chords, GraceChord, Rest
from musicscore.exceptions import ChordHasNoParentError, DeepCopyException, ChordNotesAreAlreadyCreatedError, \
    ChordException, MusicTreeException, ChordAddXPlacementException, RestCannotSetMidiError, \
    RestWithDisplayStepHasNoDisplayOctave, RestWithDisplayOctaveHasNoDisplayStep, GraceChordCannotHaveGraceNotes
from musicscore.midi import Midi
from musicscore.quarterduration import QuarterDuration
from musicscore.tests.util import ChordTestCase, create_test_objects, IdTestCase
from musicscore.util import XML_ARTICULATION_CLASSES, XML_TECHNICAL_CLASSES, XML_DYNAMIC_CLASSES, XML_ORNAMENT_CLASSES, \
    XML_OTHER_NOTATIONS, XML_DIRECTION_TYPE_CLASSES
from musicxml.xmlelement.xmlelement import *


def get_chord_midi_values(chord):
    return [m.value for m in chord._midis]


class TestTreeChord(ChordTestCase):

    def test_mocks(self):
        assert self.mock_voice.number == 1
        assert self.mock_beat.up.number == 1
        ch = Chord()
        ch._parent = self.mock_beat
        assert ch.up == self.mock_beat
        assert ch.up.up == self.mock_voice
        assert ch.up.up.up == self.mock_staff
        assert ch.up.up.up.up == self.mock_measure
        assert ch.get_parent_measure() == self.mock_measure

    def test_chord_init_midis(self):
        """
        Test all possible types of input midis
        """
        ch = Chord(70, 1)
        assert get_chord_midi_values(ch) == [70]
        ch = Chord([70, 50], 1)
        assert get_chord_midi_values(ch) == [50, 70]
        ch = Chord(0, 1)
        assert get_chord_midi_values(ch) == [0]
        ch = Chord(Midi(90, accidental=Accidental(mode='enharmonic')), 1)
        assert get_chord_midi_values(ch) == [90]
        ch = Chord([Midi(90, accidental=Accidental(mode='enharmonic')), 70], 1)
        assert get_chord_midi_values(ch) == [70, 90]

        with self.assertRaises(ValueError):
            Chord([0, 60], 1)

        """
        A grace note cannot be a rest.
        """
        with self.assertRaises(ValueError):
            Chord(0, 0)

    def test_chord_needs_parent_error(self):
        ch = Chord(70, 1)
        with self.assertRaises(ChordHasNoParentError):
            ch.finalize()
        ch._parent = self.mock_beat
        ch.finalize()

    def test_init_quarter_durations(self):
        """
        Test values of quarter_duration
        """
        ch = Chord(90, 1.25)
        assert ch.quarter_duration == 1.25
        ch = Chord(80, 1.2)
        assert ch.quarter_duration == Fraction(1.2)
        assert ch.quarter_duration == 1.2
        assert ch.quarter_duration == Fraction(6 / 5)
        assert ch.quarter_duration == 6 / 5
        ch = Chord(80, 8)
        assert ch.quarter_duration == 8
        with self.assertRaises(TypeError):
            ch = Chord(80, '1.2')
        ch = Chord(70, Fraction(1, 4))
        assert ch.quarter_duration == 0.25
        ch = Chord(70, QuarterDuration(1, 4))
        assert ch.quarter_duration == 0.25

    def test_is_rest(self):
        """
        Test is_rest property
        """
        assert Chord(0, 1).is_rest
        assert not Chord(50, 1).is_rest

    def test_chord_attributes(self):
        """
        Test that a dot operator can set and get note attributes
        """
        c = Chord([60, 61, 62], 2)
        c._parent = self.mock_beat
        c.finalize()
        c.relative_x = 10
        assert c.relative_x == [10, 10, 10]
        c.relative_y = [None, 20, 10]
        assert c.relative_y == [None, 20, 10]
        c.relative_y = [10]
        assert c.relative_y == [10, 20, 10]
        c.default_y = [10]
        assert c.default_y == [10, None, None]

    def test_chord_one_note(self):
        c = Chord(70, 4, relative_x=10)
        c.midis[0].accidental.show = True
        c._parent = self.mock_beat
        c.finalize()
        expected = """<note relative-x="10">
  <pitch>
    <step>B</step>
    <alter>-1</alter>
    <octave>4</octave>
  </pitch>
  <duration>4</duration>
  <voice>1</voice>
  <type>whole</type>
  <accidental>flat</accidental>
</note>
"""
        assert c.notes[0].to_string() == expected
        c.midis[0].value = 72
        expected = """<note relative-x="10">
  <pitch>
    <step>C</step>
    <octave>5</octave>
  </pitch>
  <duration>4</duration>
  <voice>1</voice>
  <type>whole</type>
  <accidental>natural</accidental>
</note>
"""
        assert c.notes[0].to_string() == expected
        # change chord's midi (zero)
        c.midis[0].value = 0
        # c.finalize()

        expected = """<note relative-x="10">
  <rest />
  <duration>4</duration>
  <voice>1</voice>
  <type>whole</type>
</note>
"""
        assert c.notes[0].to_string() == expected
        c = Chord(midis=0, quarter_duration=2)
        # change chord's duration (not zero)
        c.quarter_duration = 1
        c._parent = self.mock_beat
        c.finalize()
        expected = """<note>
  <rest />
  <duration>1</duration>
  <voice>1</voice>
  <type>quarter</type>
</note>
"""
        assert c.notes[0].to_string() == expected
        c = Chord(0, 1)
        # change chord's duration (zero)
        with self.assertRaises(ValueError):
            c.quarter_duration = 0
        c.midis[0].value = 60
        # c.quarter_duration = 0
        # c._parent = self.mock_beat

    def test_chord_single_non_rest(self):
        """
        Test a chord with a non rest single midi
        """
        c = Chord(72, 2)
        c.midis[0].accidental.show = True
        c._parent = self.mock_beat
        c.finalize()

        c.xml_type = '16th'
        c.xml_stem = 'up'
        c.xml_staff = 1

        expected = """<note>
  <pitch>
    <step>C</step>
    <octave>5</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>16th</type>
  <accidental>natural</accidental>
  <stem>up</stem>
  <staff>1</staff>
</note>
"""
        assert c.notes[0].to_string() == expected
        c.midis[0].value = 61
        expected = """<pitch>
  <step>C</step>
  <alter>1</alter>
  <octave>4</octave>
</pitch>
"""
        assert c.notes[0].xml_pitch.to_string() == expected
        c.midis[0].value = 0
        # c.finalize()
        assert c.notes[0].xml_pitch is None
        assert c.notes[0].xml_rest is not None
        assert c.is_rest

    def test_chord_single_non_rest_midi_with_accidental(self):
        """
        Test chord with a non rest single midi with accidental set.
        """
        chord = Chord(Midi(70, accidental=Accidental(mode='sharp')), 1)
        chord._parent = self.mock_beat
        chord.finalize()
        expected = """<pitch>
    <step>A</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
"""
        assert chord.notes[0].xml_pitch.to_string() == expected

    def test_chord_as_rest(self):
        """
        Test chord with a rest midi
        """
        chord = Chord(0, 2)
        chord._parent = self.mock_beat
        expected = """<note>
  <rest />
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
</note>
"""

        chord.finalize()
        assert chord.notes[0].to_string() == expected
        chord.midis = [60, 61]
        expected = """<pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
"""
        assert chord.notes[0].xml_pitch.to_string() == expected
        expected = """<pitch>
    <step>C</step>
    <alter>1</alter>
    <octave>4</octave>
  </pitch>
"""
        assert chord.notes[1].xml_pitch.to_string() == expected

    def test_chord_to_rest(self):
        chord = Chord(60, 2)
        chord._parent = self.mock_beat
        chord.finalize()
        chord.to_rest()
        expected = """<note>
  <rest />
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
</note>
"""
        assert chord.notes[0].to_string() == expected
        assert len(chord.notes) == 1

    def test_chord_with_multiple_midis(self):
        """
        Test chord with a list of midis
        """

        chord = Chord([60, 62, 63], 2)
        chord.midis[1].accidental.show = True
        chord.midis[2].accidental.show = True
        chord._parent = self.mock_beat
        chord.finalize()
        chord.xml_stem = 'up'
        expected_1 = """<note>
  <pitch>
    <step>C</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <stem>up</stem>
</note>
"""
        expected_2 = """<note>
  <chord />
  <pitch>
    <step>D</step>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <accidental>natural</accidental>
  <stem>up</stem>
</note>
"""
        expected_3 = """<note>
  <chord />
  <pitch>
    <step>E</step>
    <alter>-1</alter>
    <octave>4</octave>
  </pitch>
  <duration>2</duration>
  <voice>1</voice>
  <type>half</type>
  <accidental>flat</accidental>
  <stem>up</stem>
</note>
"""
        for note, expected in zip(chord.notes, [expected_1, expected_2, expected_3]):
            assert note.to_string() == expected

    def test_group_chords(self):
        chords = [Chord(60, qd) for qd in [1 / 6, 1 / 6, 1 / 6, 1 / 10, 3 / 10, 1 / 10]]
        with self.assertRaises(ValueError):
            assert _group_chords(chords, [1 / 2, 1 / 2, 1 / 2])
        assert _group_chords(chords, [1 / 2, 1 / 2]) == [chords[:3], chords[3:]]
        assert _group_chords(chords, [1 / 3, 2 / 3]) == [chords[:2], chords[2:]]
        assert _group_chords(chords, [1 / 4, 3 / 4]) is None

    def test_has_same_pitches(self):
        ch1 = Chord([60, Midi(61), 62], 1)
        ch2 = Chord([60, Midi(61)], 1)
        assert not ch1.has_same_pitches(ch2)
        ch2 = Chord([60, Midi(61), 62], 1)
        assert ch1.has_same_pitches(ch2)
        ch2 = Chord([60, Midi(61, accidental=Accidental(mode='flat')), 62], 1)
        assert not ch1.has_same_pitches(ch2)

    def test_add_lyric(self):
        ch = Chord(60, 2)
        ch._parent = self.mock_beat
        ch.add_lyric('test')
        ch.finalize()
        assert ch.notes[0].xml_lyric is not None
        assert ch.notes[0].xml_lyric.xml_text.value_ == 'test'

    def test_add_lyrics_after_creating_notes(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        lyrics1 = ch.add_lyric('one')
        ch.finalize()
        lyrics2 = ch.add_lyric('two')
        assert ch.notes[0].find_children('XMLLyric') == [lyrics1, lyrics2]

    def test_get_staff_number(self):
        ch = Chord(60, 2)
        ch._parent = self.mock_beat
        assert ch.get_staff_number() is None
        self.mock_staff.number = 1
        assert ch.get_staff_number() == 1

    def test_add_clef(self):
        ch = Chord(60, 2)
        assert ch.clef is None
        cl = BassClef()
        ch.clef = cl
        assert ch.clef == cl
        with self.assertRaises(TypeError):
            ch.clef = 'bla'

    @skip
    def test_percussion_notation(self):
        self.fail('Incomplete')

    @skip
    def test_finger_tremolo(self):
        self.fail('Incomplete')

    def test_deepcopy_chord(self):
        chord = Chord([60, 62], 2)
        chord.add_tie('start')
        copied = copy.deepcopy(chord)
        assert [midi.value for midi in copied.midis] == [midi.value for midi in chord.midis]
        assert [id(midi) for midi in copied.midis] != [id(midi) for midi in chord.midis]
        assert chord.quarter_duration.value == copied.quarter_duration.value
        assert id(chord.quarter_duration) != id(copied.quarter_duration)
        chord._parent = self.mock_beat
        chord.finalize()
        with self.assertRaises(DeepCopyException):
            copy.deepcopy(chord)
        # chord.add_dynamics('p')
        # chord.add_lyric('something')
        # chord.add_x(XMLAccent())
        # chord.add_x(XMLUpBow())
        # chord.add_x(XMLAccidentalMark())
        # chord.add_x(XMLFf())

    def test_add_midi(self):
        chord = Chord([60, 62], 2)
        m = Midi(63)
        chord.add_midi(m)
        assert [midi.value for midi in chord.midis] == [60, 62, 63]
        assert chord.midis[-1] == m
        chord.add_midi(58)
        assert [midi.value for midi in chord.midis] == [58, 60, 62, 63]
        chord.add_midi(60)
        assert [midi.value for midi in chord.midis] == [58, 60, 60, 62, 63]

        chord._parent = self.mock_beat
        chord.finalize()
        with self.assertRaises(ChordNotesAreAlreadyCreatedError):
            chord.add_midi(80)

    def test_add_direction_type(self):
        score = Score()
        p = score.add_part('part-1')
        for dt_class in XML_DIRECTION_TYPE_CLASSES:
            chord = Chord(midis=60, quarter_duration=4)
            if dt_class == XMLSymbol:
                dt_obj = dt_class('0')
            else:
                dt_obj = dt_class()
            if dt_class == XMLDynamics:
                assert self.assertRaises(ChordException)
            else:
                chord.add_direction_type(dt_obj, 'above')
                assert dt_obj in chord.xml_direction_types['above']
                assert chord.xml_direction_types['above'] == [dt_obj]
                p.add_chord(chord)
        p.finalize()

    def test_add_direction_type_wrong_type(self):
        chord = Chord(60, 4)
        with self.assertRaises(TypeError):
            chord.add_direction_type(XMLFermata())

    def test_chord_to_string(self):
        p = Part('p1')
        ch = Chord(60, 1)
        p.add_chord(ch)
        ch.finalize()
        with self.assertRaises(MusicTreeException):
            ch.to_string()

    def test_get_x(self):
        def test_get_objects(*classes):
            chord = Chord(60, 1)
            cl1, cl2, cl3 = classes
            dt1 = chord.add_x(cl1())
            dt2 = chord.add_x(cl1())
            dt3 = chord.add_x(cl2())
            assert chord.get_x(cl1) == [dt1, dt2]
            assert chord.get_x(cl2) == [dt3]
            assert chord.get_x(cl3) == []

        # oranment
        test_get_objects(XML_ORNAMENT_CLASSES[0], XML_ORNAMENT_CLASSES[1], XML_ORNAMENT_CLASSES[2])
        # articualtion
        test_get_objects(XML_ARTICULATION_CLASSES[0], XML_ARTICULATION_CLASSES[1], XML_ARTICULATION_CLASSES[2])
        # technical
        test_get_objects(XML_TECHNICAL_CLASSES[0], XML_TECHNICAL_CLASSES[1], XML_TECHNICAL_CLASSES[2])
        # other notations
        test_get_objects(XML_OTHER_NOTATIONS[0], XML_OTHER_NOTATIONS[1], XML_OTHER_NOTATIONS[2])
        # direction type
        test_get_objects(XML_DIRECTION_TYPE_CLASSES[0], XML_DIRECTION_TYPE_CLASSES[1], XML_DIRECTION_TYPE_CLASSES[2])

        # ornament or notation
        chord = Chord(60, 1)
        cl1 = XMLAccidentalMark
        dt1 = chord.add_x(cl1('sharp'), parent_type='ornament')
        dt2 = chord.add_x(cl1('flat'), parent_type='ornament')
        dt3 = chord.add_x(cl1('flat'), parent_type='notation')
        assert chord.get_x(cl1) == [dt1, dt2, dt3]

    def test_get_wedges(self):
        chord = Chord(60, 1)
        dt1 = chord.add_x(XMLWedge(type='crescendo'))
        dt2 = chord.add_x(XMLWedge(type='continue'))
        dt3 = chord.add_x(XMLWedge(type='stop'))
        assert chord.get_wedges() == [dt1, dt2, dt3]

    def test_get_slurs(self):
        chord = Chord(60, 1)
        dt1 = chord.add_x(XMLSlur(type='start'))
        dt2 = chord.add_x(XMLSlur(type='continue'))
        dt3 = chord.add_x(XMLSlur(type='stop'))
        assert chord.get_slurs() == [dt1, dt2, dt3]

    def test_get_words(self):
        chord = Chord(60, 1)
        dt1 = chord.add_x(XMLWords('something'))
        dt2 = chord.add_x(XMLWords('something else'))
        assert chord.get_words() == [dt1, dt2]

    def test_get_brackets(self):
        chords = [Chord(60, 1) for _ in range(3)]
        b1, b2, b3 = chords[0].add_x(XMLBracket()), chords[1].add_x(XMLBracket()), chords[2].add_x(XMLBracket())
        assert [ch.get_brackets()[0] for ch in chords] == [b1, b2, b3]

    def test_add_arpeggio_normal(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = 'normal'
        assert ch.arpeggio == 'normal'
        ch.finalize()
        for n in ch.notes:
            assert n.xml_notations.xml_arpeggiate
            assert n.xml_notations.xml_arpeggiate.direction is None

    def test_add_arpeggio_up(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = 'up'
        assert ch.arpeggio == 'up'
        ch.finalize()
        for n in ch.notes:
            assert n.xml_notations.xml_arpeggiate
            if n == ch.notes[0]:
                assert n.xml_notations.xml_arpeggiate.direction == 'up'

    def test_add_arpeggio_bottom(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = 'down'
        assert ch.arpeggio == 'down'
        ch.finalize()
        for n in ch.notes:
            assert n.xml_notations.xml_arpeggiate
            if n == ch.notes[0]:
                assert n.xml_notations.xml_arpeggiate.direction == 'down'

    def test_add_none_arpeggio(self):
        ch = Chord([60, 64, 67], 1)
        assert not ch.arpeggio
        ch._parent = self.mock_beat
        ch.arpeggio = 'none'
        assert ch.arpeggio == 'none'
        ch.finalize()
        for n in ch.notes:
            if n == ch.notes[0]:
                assert n.xml_notations.xml_non_arpeggiate
                assert n.xml_notations.xml_non_arpeggiate.type == 'bottom'
            elif n == ch.notes[-1]:
                assert n.xml_notations.xml_non_arpeggiate
                assert n.xml_notations.xml_non_arpeggiate.type == 'top'


class TestTreeRest(ChordTestCase):
    def test_rest_init(self):
        r = Rest(4)
        r._parent = self.mock_beat
        assert isinstance(r, Rest)
        assert isinstance(r, Chord)
        assert r.midis[0].value == 0
        r.finalize()
        n = r.notes[0]
        assert r.display_step == r.display_octave == n.xml_object.xml_rest.xml_display_step == n.xml_object.xml_rest.xml_display_octave is None
        assert r.quarter_duration == 4
        r = Rest(4, display_step='C', display_octave=4)
        r._parent = self.mock_beat
        r.finalize()
        n = r.notes[0]
        assert r.display_step == n.xml_object.xml_rest.xml_display_step.value_ == 'C'
        assert r.display_octave == n.xml_object.xml_rest.xml_display_octave.value_ == 4
        r = Rest(4, measure='yes')
        assert r.measure == 'yes'
        r._parent = self.mock_beat
        r.finalize()
        n = r.notes[0]
        assert n.xml_object.xml_rest.measure == 'yes'
        with self.assertRaises(RestCannotSetMidiError):
            Rest(midis=0, quarter_duration=2)
        with self.assertRaises(TypeError):
            Rest()
        with self.assertRaises(RestWithDisplayStepHasNoDisplayOctave):
            r = Rest(4, display_step='D')
            r._parent = self.mock_beat
            r.finalize()
        with self.assertRaises(RestWithDisplayOctaveHasNoDisplayStep):
            r = Rest(4, display_octave=4)
            r._parent = self.mock_beat
            r.finalize()
        with self.assertRaises(TypeError):
            Rest(4, display_step='H', display_octave=2)
        with self.assertRaises(TypeError):
            Rest(4, display_step='C', display_octave=2.2)
        with self.assertRaises(TypeError):
            Rest(4, display_step='C', display_octave=-2)


class TestTies(ChordTestCase):

    def test_add_tie_ties_midis(self):
        ch = Chord(midis=[60, 63])
        ch._parent = self.mock_beat
        ch.add_tie('start')
        for midi in ch.midis:
            assert midi._ties == {'start'}
        ch.finalize()
        assert [n.is_tied for n in ch.notes] == [True, True]

    def test_tied_midis(self):
        m1 = Midi(60)
        m1.add_tie('start')
        m2 = Midi(61)
        m2.add_tie('start')
        ch = Chord(midis=[m1, m2])
        ch._parent = self.mock_beat
        ch.finalize()
        assert [n.is_tied for n in ch.notes] == [True, True]

    def test_split_tied_copy(self):
        ch = Chord(midis=60, quarter_duration=1)
        copied = _split_copy(ch)
        assert ch.midis[0]._ties == copied.midis[0]._ties == set()

    def test_tie_one_note(self):
        ch1 = Chord(midis=[60, 63])
        ch2 = Chord(midis=[60, 65])
        ch1.midis[0].add_tie('start')
        ch2.midis[0].add_tie('stop')
        ch1._parent = self.mock_beat
        ch2._parent = self.mock_beat
        ch1.finalize()
        ch2.finalize()
        assert [n.is_tied for n in ch1.notes] == [True, False]
        assert [n.is_tied_to_previous for n in ch2.notes] == [True, False]

    def test_untie_one_note(self):
        ch = Chord(midis=[60, 61])
        ch.add_tie('start')
        ch._parent = self.mock_beat
        ch.finalize()
        assert [n.is_tied for n in ch.notes] == [True, True]
        # print(ch.notes)
        ch.midis[0].remove_tie('start')
        assert [n.is_tied for n in ch.notes] == [False, True]

    def test_chord_tie_untie(self):
        ch1 = Chord(midis=[60, 61])
        ch2 = Chord(midis=[60, 61])
        ch1.add_tie('start')
        ch2.add_tie('stop')
        ch1._parent = self.mock_beat
        ch2._parent = self.mock_beat
        ch1.finalize()
        ch2.finalize()
        assert [n.is_tied for n in ch1.notes] == [True, True]
        assert [n.is_tied for n in ch2.notes] == [False, False]
        assert [n.is_tied_to_previous for n in ch2.notes] == [True, True]

    def test_chord_change_tie_after_finalizing(self):
        ch1 = Chord(midis=[60, 61])
        ch2 = Chord(midis=[60, 61])
        ch1._parent = self.mock_beat
        ch2._parent = self.mock_beat
        ch1.finalize()
        ch2.finalize()

        ch1.add_tie('start')
        ch2.midis[1].add_tie('stop')

        assert [n.is_tied for n in ch1.notes] == [True, True]
        assert [n.is_tied for n in ch2.notes] == [False, False]
        assert [n.is_tied_to_previous for n in ch2.notes] == [False, True]

    def test_chord_all_midis_tied_to_next_or_previous(self):
        ch = Chord([60, 61, 62])
        for m in ch.midis:
            m.add_tie('start')
        assert ch.all_midis_are_tied_to_next
        ch.midis[1].remove_tie('start')
        assert not ch.all_midis_are_tied_to_next
        ch.midis[0].add_tie('stop')
        assert not ch.all_midis_are_tied_to_previous
        for m in ch.midis:
            m.add_tie('stop')
        assert ch.all_midis_are_tied_to_previous


class TestSplit(TestCase):

    def test_set_original_starting_ties(self):
        ch = Chord(midis=[60, 61], quarter_duration=1)
        ch._set_original_starting_ties(ch)
        copied = _split_copy(ch)
        assert copied._original_starting_ties == [set(), set()]
        ch.midis[0].add_tie('start')
        ch._set_original_starting_ties(ch)
        assert copied._original_starting_ties == [set(), set()]

    def test_split_copy(self):
        ch = Chord(midis=[Midi(61, accidental=Accidental(mode='sharp'))], quarter_duration=2, offset=0.5)
        copied = _split_copy(ch)
        assert ch.midis != copied.midis
        assert ch.midis[0].value == copied.midis[0].value
        assert ch.midis[0].accidental != copied.midis[0].accidental
        assert ch.midis[0].accidental.mode == copied.midis[0].accidental.mode
        copied.midis[0].accidental.show = False
        assert ch.midis[0].accidental.show is None
        assert copied.midis[0].accidental.show is False

    def test_tied_split_copy(self):
        ch = Chord(midis=61, quarter_duration=2)
        ch.add_tie('start')
        copied = _split_copy(ch)
        for m in copied.midis:
            assert m._ties == set()

    def test_split_quarter_durations(self):
        ch = Chord(midis=60, quarter_duration=4)
        copied = _split_copy(ch)
        assert id(ch.quarter_duration) != id(copied.quarter_duration)
        ch.quarter_duration = 2
        assert copied.quarter_duration == 4
        assert ch.quarter_duration == 2

        ch = Chord(midis=60, quarter_duration=5)
        beats = [Beat(1), Beat(1), Beat(1), Beat(1)]
        quarter_durations = ch.quarter_duration.get_beatwise_sections(beats=beats)
        ch.quarter_duration = quarter_durations[0][0]
        copied = _split_copy(ch, quarter_durations[1])
        assert [ch.quarter_duration, copied.quarter_duration] == [4, 1]


class TestAddGraceChord(ChordTestCase):
    def test_add_grace_chord_parameters(self):
        ch = Chord(60, 1)
        g1 = gch = ch.add_grace_chord(midis_or_grace_chord=60)
        assert gch.midis[0].value == 60
        in_gch = GraceChord(61)
        g2 = gch = ch.add_grace_chord(midis_or_grace_chord=in_gch)
        assert gch == in_gch
        g3 = gch = ch.add_grace_chord(midis_or_grace_chord=[62, 63])
        assert [m.value for m in gch.midis] == [62, 63]
        g4 = gch = ch.add_grace_chord(64)
        assert gch.midis[0].value == 64
        g5 = gch = ch.add_grace_chord(midis_or_grace_chord=65, type_='16th')
        assert gch.midis[0].value == 65
        assert gch.type_.value_ == '16th'
        in_gch = GraceChord(61, type_='16th')
        g6 = gch = ch.add_grace_chord(midis_or_grace_chord=in_gch)
        assert gch == in_gch
        assert gch.type_.value_ == '16th'
        with self.assertRaises(ValueError):
            ch.add_grace_chord(midis_or_grace_chord=GraceChord(66, type_='quarter'), type_='16th')
        g7 = gch = ch.add_grace_chord(67, type_='16th')
        assert gch.midis[0].value == 67
        assert gch.type_.value_ == '16th'
        g8 = gch = ch.add_grace_chord(68, type_='16th', position='after')
        assert gch.position == 'after'
        assert gch.midis[0].value == 68
        assert gch.type_.value_ == '16th'
        with self.assertRaises(ValueError):
            ch.add_grace_chord(GraceChord(66, type_='quarter'), position='after')
        assert ch.get_grace_chords(position='before') == [g1, g2, g3, g4, g5, g6, g7]
        assert ch.get_grace_chords(position='after') == [g8]

    def test_add_grace_chord_before_and_after_argument(self):
        ch = Chord(60, 1)
        midis_or_chord = [62, [63, 65], GraceChord(66)]
        before_chords = [ch.add_grace_chord(x) for x in midis_or_chord]
        assert ch._grace_chords['before'] == before_chords

        midis_or_chord = [67, [68, 70], GraceChord(71, position='after')]
        after_chords = [
            ch.add_grace_chord(x, position='after') if not isinstance(x, GraceChord) else ch.add_grace_chord(x) for x in
            midis_or_chord]
        assert ch._grace_chords['after'] == after_chords

    def test_add_grace_chord_generates_grace_chords(self):
        ch = Chord(60, 1)
        with self.assertRaises(TypeError):
            ch.add_grace_chord()
        gc1 = ch.add_grace_chord(60)
        gc2 = ch.add_grace_chord(61, type_='quarter')
        gc3 = ch.add_grace_chord(GraceChord(midis=[62, 64]))
        gc4 = ch.add_grace_chord(GraceChord(63, type_='16th'))
        all_gcs = [gc1, gc2, gc3, gc4]
        assert [[m.value for m in gc.midis] for gc in all_gcs] == [[60], [61], [62, 64], [63]]
        assert [gc.type_.value_ if gc.type_ else None for gc in all_gcs] == [None, 'quarter', None, '16th']

    def test_add_grace_chord_finalize(self):
        part = Part('p1')
        ch = Chord(midis=[60, 63], quarter_duration=4)
        midis_or_chord = [62, [63, 65], GraceChord(66)]
        before_chords = [ch.add_grace_chord(x) for x in midis_or_chord]
        midis_or_chord = [67, [68, 70], GraceChord(71, position='after')]
        after_chords = [
            ch.add_grace_chord(x, position='after') if not isinstance(x, GraceChord) else ch.add_grace_chord(x) for x in
            midis_or_chord]
        part.add_chord(ch)
        with self.assertRaises(ChordException):
            ch.add_grace_chord(80)
        part.finalize()
        assert len(part.get_chords()) == 7

    def test_grace_chords_are_added_to_right_voice(self):
        part = Part('p1')
        ch = Chord(midis=[60, 63], quarter_duration=4)
        gch1 = ch.add_grace_chord(80)
        gch2 = ch.add_grace_chord(90, position='after')
        part.add_chord(ch, staff_number=2, voice_number=2)
        part.finalize()
        assert part.get_chords() == [gch1, ch, gch2]
        for c in [gch1, ch, gch2]:
            assert c.voice_number == 2
            assert c.get_staff_number() == 2

    def test_grace_chords_after_the_last_chord_in_measure(self):
        part = Part('p1')
        ch = Chord(60, 4)
        gch = ch.add_grace_chord(80, position='after')
        part.add_chord(ch)
        part.add_chord(Chord(90, 1))
        part.finalize()
        assert gch in part.get_measure(1).get_chords()

    def test_error_grace_chord_add_grace_chord(self):
        gch = GraceChord(60)
        with self.assertRaises(GraceChordCannotHaveGraceNotes):
            gch.add_grace_chord(80)

    def test_grace_chords_after_to_right_beat(self):
        part = Part('p1')
        ch = Chord(60, 1)
        gch = ch.add_grace_chord(80, position='after')
        part.add_chord(ch)
        part.add_chord(Chord(90, 3))
        assert gch in part.get_beat(1, 1, 1, 1).get_children()


class TestAddX(ChordTestCase):
    # def setUp(self):
    #     super().setUp()
    #     self.score = Score()
    #     self.part = self.score.add_part('p1')

    def test_add_x_parent_type(self):
        ch = Chord(60, 1)
        ch.add_x(XML_DIRECTION_TYPE_CLASSES[0]())

    def test_add_x_placement(self):
        def add_type_and_placement(type_, xml_object=None):
            above_part = Part(f"above-{type_}-{xml_object.name}") if xml_object else Part(f"above-{type_}")
            below_part = Part(f"below-{type_}-{xml_object.name}") if xml_object else Part(f"below-{type_}")
            if not xml_object:
                above_objects = create_test_objects(type_)
                below_objects = create_test_objects(type_)
            else:
                above_objects = [xml_object]
                below_objects = [xml_object.__deepcopy__()]

            for obj in above_objects:
                ch = Chord(60, 1)
                try:
                    ch.add_x(obj, placement='above', parent_type=type_)
                except ChordAddXPlacementException:
                    ch.add_x(obj, placement=None, parent_type=type_)
                above_part.add_chord(ch)
            for obj in below_objects:
                ch = Chord(60, 1)
                try:
                    ch.add_x(obj, placement='below', parent_type=type_)
                except ChordAddXPlacementException:
                    ch.add_x(obj, placement=None, parent_type=type_)
                below_part.add_chord(ch)

            above_part.finalize()
            below_part.finalize()
            if type_ == 'direction_type':
                for p in ['above', 'below']:
                    directions = [d for m in eval(f"{p}_part.get_children()") for d in
                                  [c for c in m.xml_object.get_children() if isinstance(c, XMLDirection)]]
                    assert len(directions) == len(above_objects)
                    for x in directions:
                        assert x.placement == p

            elif type_ in ['technical', 'ornament', 'articulation']:
                for p in ['above', 'below']:
                    notations = [notation for measure in eval(f'{p}_part').get_children() for note in
                                 [x for x in measure.xml_object.get_children() if isinstance(x, XMLNote)] for notation
                                 in
                                 [y for y in note.get_children() if isinstance(y, XMLNotations)]]
                    assert len(notations) == len(above_objects)
                    for x in notations:
                        if x.get_children()[0].get_children()[0].__class__ not in [XMLFret, XMLBend]:
                            assert x.get_children()[0].get_children()[0].placement == p
            else:
                raise NotImplementedError(f'testing type_ {type_}')

        # direction_type objects
        add_type_and_placement('direction_type')

        # technical objects
        add_type_and_placement('technical')

        # ornament objects
        add_type_and_placement('ornament')

        # articulation objects
        add_type_and_placement('articulation')

        # notation only objects does not accept placement (except fermata)
        notation_object = XMLArpeggiate()
        ch = Chord(60, 1)
        with self.assertRaises(ChordAddXPlacementException):
            ch.add_x(notation_object, placement='above', parent_type='notation')
        ch.add_x(notation_object, placement=None, parent_type='notation')

        # mixed below and above
        part = Part('mixed_placements')
        x_objects = [XMLTenuto(), XMLStaccato(), XMLAccent()]
        placements = ['below', 'below', 'above']
        ch = Chord(60, 1)
        for obj, plc in zip(x_objects, placements):
            ch.add_x(obj, placement=plc)
        part.add_chord(ch)
        part.finalize()
        notations = [notation for measure in part.get_children() for note in
                     [x for x in measure.xml_object.get_children() if isinstance(x, XMLNote)] for notation in
                     [y for y in note.get_children() if isinstance(y, XMLNotations)]]
        for index, x in enumerate(notations[0].xml_articulations.get_children()):
            assert x.placement == placements[index]

        # fermata placement is translated in fermata types inverted or upright
        fermatas = [XMLFermata(), XMLFermata()]
        placements = ['above', 'below']
        ch = Chord(60, 1)
        for f, p in zip(fermatas, placements):
            ch.add_x(f, placement=p)

        for f, t in zip(fermatas, ['upright', 'inverted']):
            assert f.type == t

    def test_add_articulation_after_creating_notes(self):
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        staccato = ch.add_x(XMLStaccato())
        ch.finalize()
        assert ch.notes[0].xml_notations.xml_articulations.get_children() == [staccato]
        assert isinstance(ch.notes[0].xml_notations.xml_articulations.get_children()[0], XMLStaccato)
        accent = ch.add_x(XMLAccent())
        assert ch.notes[0].xml_notations.xml_articulations.get_children() == [staccato, accent]

    def test_add_multiple_articulations(self):
        articulation_classes = XML_ARTICULATION_CLASSES[:3]
        ch = Chord(60, 1)
        ch._parent = self.mock_beat
        for a in articulation_classes:
            ch.add_x(a())
        assert len(ch._xml_articulations) == 3
        ch.finalize()
        n = ch.notes[0]
        assert n.xml_notations.xml_articulations is not None
        assert [type(a) for a in n.xml_notations.xml_articulations.get_children()] == articulation_classes

    def test_chord_add_x_as_object_articulation(self):
        for obj in create_test_objects('articulation'):
            ch = Chord(60, 1)
            ch.add_x(obj)
            ch._parent = self.mock_beat
            ch.finalize()
            assert ch.notes[0].xml_notations.xml_articulations.get_children()[0] == obj

    def test_chord_add_x_as_object_technical(self):
        for obj in create_test_objects('technical'):
            ch = Chord(60, 1)
            ch.add_x(obj)
            ch._parent = self.mock_beat
            ch.finalize()
            assert ch.notes[0].xml_notations.xml_technical.get_children()[0] == obj

    def test_chord_add_x_as_object_dynamics(self):
        for cls in XML_DYNAMIC_CLASSES:
            ch = Chord(60, 1)
            ch.add_x(cls(), parent_type='notation')
            ch._parent = self.mock_beat
            ch.finalize()
            assert isinstance(ch.notes[0].xml_notations.xml_dynamics.get_children()[0], cls)

    def test_chord_add_x_as_object_ornaments(self):
        for obj in create_test_objects('ornament'):
            ch = Chord(60, 1)
            if isinstance(obj, XMLAccidentalMark):
                ch.add_x(obj, parent_type='ornament')
            else:
                ch.add_x(obj)
            ch._parent = self.mock_beat
            ch.finalize()
            assert ch.notes[0].xml_notations.xml_ornaments.get_children()[0] == obj

    def test_chord_add_x_trill_with_wavy_line_and_accidental_mark(self):
        ch = Chord(60, 1)
        ch.add_x(XMLTrillMark())
        ch.add_x(XMLAccidentalMark('sharp'), parent_type='ornament')
        ch.add_x(XMLWavyLine(type='start', relative_x=0))
        ch.add_x(XMLWavyLine(type='stop', relative_x=20))
        ch._parent = self.mock_beat
        ch.finalize()
        expected = """<ornaments>
      <trill-mark />
      <accidental-mark>sharp</accidental-mark>
      <wavy-line type="start" relative-x="0" />
      <wavy-line type="stop" relative-x="20" />
    </ornaments>
"""
        assert ch.notes[0].xml_notations.xml_ornaments.to_string() == expected

    def test_chord_add_x_as_object_other_notations(self):
        for cls in XML_OTHER_NOTATIONS:
            ch = Chord(60, 1)
            ch.add_x(cls())
            ch._parent = self.mock_beat
            ch.finalize()
            assert isinstance(ch.notes[0].xml_notations.get_children()[0], cls)

    def test_add_xml_wedge_objects(self):
        wedges = [XMLWedge(type=val) for val in ['crescendo', 'stop', 'diminuendo', 'stop']]
        for wedge in wedges:
            ch = Chord(60, 4)
            ch._parent = self.mock_beat
            ch.add_wedge(wedge)
            ch.finalize()
            assert len(ch._xml_directions) == 1
            d = ch._xml_directions[0]
            assert d.placement == 'below'
            assert d.xml_direction_type.xml_wedge == wedge

    def test_add_wedge_string(self):
        wedges = ['crescendo', 'stop', 'diminuendo', 'stop']
        for wedge in wedges:
            ch = Chord(60, 4)
            ch._parent = self.mock_beat
            ch.add_wedge(wedge)
            ch.finalize()
            assert len(ch._xml_directions) == 1
            d = ch._xml_directions[0]
            assert d.placement == 'below'
            assert d.xml_direction_type.xml_wedge.type == wedge

    def test_add_words(self):
        ch = Chord(60, 4)
        ch._parent = self.mock_beat
        texts = ['Below', 'Above', 'FontSize', 'Bold']
        xml_words_objects = [ch.add_words(texts[0], placement='below'),
                             ch.add_words(texts[1]),
                             ch.add_words(texts[2], font_size=18, relative_y=30),
                             ch.add_words(texts[3], font_weight='bold', relative_y=60)]
        ch.finalize()
        assert set(xml_words_objects) == set([d.xml_direction_type.xml_words for d in ch._xml_directions])

        for obj, t in zip(xml_words_objects, texts):
            assert obj.value_ == t

        for obj, p in zip(xml_words_objects, ['below', 'above', 'above', 'above']):
            assert obj.up.up.placement == p
        for fs, obj in zip([None, None, 18, None], xml_words_objects):
            assert obj.font_size == fs
        for fw, obj in zip([None, None, None, 'bold'], xml_words_objects):
            assert obj.font_weight == fw
        for ry, obj in zip([None, None, 30, 60], xml_words_objects):
            assert obj.relative_y == ry

        ch = Chord(60, 4)
        ch._parent = self.mock_beat

        xml_words = ch.add_words(XMLWords('something', font_size=10, relative_y=20), placement='below', font_size=20)
        ch.finalize()
        d = ch._xml_directions[0]
        assert d.xml_direction_type.xml_words.value_ == 'something'
        assert d.placement == 'below'
        assert xml_words.font_size == 20
        assert xml_words.relative_y == 20


class TestAddAfterNotes(IdTestCase):
    def test_add_after_notes(self):
        part = Part('p1')
        chords = [Chord([60, 62], 2), Chord(64, 2)]
        chords[0].add_dynamics('ff')
        [part.add_chord(ch) for ch in chords]
        b = XMLBarline(location='middle')
        b.xml_bar_style = 'dashed'
        chords[0].add_after_note_xml_objects(b)
        part.finalize()
        m = part.get_measure(1)
        expected = """<measure number="1">
    <attributes>
      <divisions>1</divisions>
      <key>
        <fifths>0</fifths>
      </key>
      <time>
        <beats>4</beats>
        <beat-type>4</beat-type>
      </time>
      <clef>
        <sign>G</sign>
        <line>2</line>
      </clef>
    </attributes>
    <direction placement="below">
      <direction-type>
        <dynamics>
          <ff />
        </dynamics>
      </direction-type>
    </direction>
    <note>
      <pitch>
        <step>C</step>
        <octave>4</octave>
      </pitch>
      <duration>2</duration>
      <voice>1</voice>
      <type>half</type>
    </note>
    <note>
      <chord />
      <pitch>
        <step>D</step>
        <octave>4</octave>
      </pitch>
      <duration>2</duration>
      <voice>1</voice>
      <type>half</type>
    </note>
    <barline location="middle">
      <bar-style>dashed</bar-style>
    </barline>
    <note>
      <pitch>
        <step>E</step>
        <octave>4</octave>
      </pitch>
      <duration>2</duration>
      <voice>1</voice>
      <type>half</type>
    </note>
  </measure>
"""
        assert m.to_string() == expected



"""Lyrics spanners: continued syllables and extenders, possibly spanning multiple notes. The intermediate notes do
not have any <lyric> element."""

from pathlib import Path

from musicscore import Score, Chord
from musicscore.lyrics import Lyrics
from musicscore.tests.util import IdTestCase


class TestLily61k(IdTestCase):
    def test_lily_61h_Lyrics_SprannersExtenders(self):
        score = Score()
        part = score.add_part('p1')

        chords = [Chord(72, 1) for _ in range(11)]
        chords.append(Chord(0, 1))

        Lyrics([('A', None), ('b', None, None, 'CC', None, None), ('e', None, None)]).add_to_chords(chords[:-1])

        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from typing import Optional

from musicscore.xmlwrapper import XMLWrapper
from musicxml.xmlelement.xmlelement import XMLClef

__all__ = ['Clef', 'TrebleClef', 'BassClef', 'AltoClef', 'TenorClef']


class Clef(XMLWrapper):
    _ATTRIBUTES = {'show', 'sign', 'line', 'octave_change'}
    XMLClass = XMLClef

    def __init__(self, sign: str = 'G', line: Optional[int] = 2, octave_change: int = None, show: bool = True,
                 default: bool = False, *args, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._show = None
        self.show = show
        self.line = line
        self.sign = sign
        self.octave_change = octave_change

        self._default = default

    @property
    def line(self) -> Optional[int]:
        """
        Sets and gets ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLLine`

        :return: ``self.xml_object.xml_line.value_``
        :rtype: int, None
        """
        if self.xml_object.xml_line:
            return self.xml_object.xml_line.value_

    @line.setter
    def line(self, val):
        self.xml_object.xml_line = val

    @property
    def octave_change(self) -> Optional[int]:
        """
        Sets and gets ``value_`` (int) of :obj:`~musicxml.xmlelement.xmlelement.XMLClefOctaveChange`and indicates how
        many octaves to add to get from written pitch to sounding pitch.
        :return: ``self.xml_object.xml_clef_octave_change.value_``
        :rtype: int, None
        """

        if self.xml_object.xml_clef_octave_change:
            return self.xml_object.xml_clef_octave_change.value_

    @octave_change.setter
    def octave_change(self, val):
        self.xml_object.xml_clef_octave_change = val

    @property
    def sign(self) -> Optional[str]:
        """
        Sets and gets ``value_`` of :obj:`~musicxml.xmlelement.xmlelement.XMLSign`

        :return: ``self.xml_object.xml_sign.value_``
        :rtype: str, None
        """
        if self.xml_object.xml_sign:
            return self.xml_object.xml_sign.value_

    @sign.setter
    def sign(self, val):
        self.xml_object.xml_sign = val

    @property
    def show(self) -> bool:
        """
        Sets and gets show attribute. If ``False`` :obj:`Clef` element is not shown.
        """
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError
        self._show = val

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def __copy__(self):
        new_clef = self.__class__()
        new_clef.sign = self.sign
        new_clef.line = self.line
        new_clef.octave_change = self.octave_change
        new_clef.show = self._show
        return new_clef


class TrebleClef(Clef):
    """
    Default parameters:
      - sign='G'
      - line=2
    """

    def __init__(self, show: bool = True, octave_change: Optional[int] = None, **kwargs):
        super().__init__(sign='G', line=2, show=show, octave_change=octave_change, **kwargs)


class BassClef(Clef):
    """
    Default parameters:
      - sign='F'
      - line=4
    """

    def __init__(self, show: bool = True, octave_change: Optional[int] = None, **kwargs):
        super().__init__(sign='F', line=4, show=show, octave_change=octave_change, **kwargs)


class AltoClef(Clef):
    """
    Default parameters:
      - sign='C'
      - line=3
    """

    def __init__(self, show: bool = True, octave_change: Optional[int] = None, **kwargs):
        super().__init__(sign='C', line=3, show=show, octave_change=octave_change, **kwargs)


class TenorClef(Clef):
    """
    Default parameters:
      - sign='C'
      - line=4
    """

    def __init__(self, show: bool = True, octave_change: Optional[int] = None, **kwargs):
        super().__init__(sign='C', line=4, show=show, octave_change=octave_change, **kwargs)


class PercussionClef(Clef):
    """
    Default parameters:
      - sign=percussion
      - line=None
    """

    def __init__(self, show: bool = True, octave_change: Optional[int] = None, **kwargs):
        super().__init__(sign='percussion', line=None, show=show, octave_change=octave_change, **kwargs)



"""
Different types of (non-repeat) barlines: default (no setting), regular, dotted, dashed, heavy, light-light, light-heavy,
heavy-light, heavy-heavy, tick, short, none.
"""
from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase
from musicxml.xmlelement.xmlelement import XMLWords


class TestLily46a(IdTestCase):
    def test_lily_46a_Barlines(self):
        score = Score()
        p = score.add_part('p1')
        p.add_chord(Chord(0, 4))

        barline_styles = ['regular', 'dotted', 'dashed', 'heavy', 'light-light', 'light-heavy',
                          'heavy-light', 'heavy-heavy', 'tick', 'short', 'none']

        for bs in barline_styles:
            ch = Chord(0, 4)
            ch.add_x(XMLWords(bs), placement='below')
            p.add_chord(ch)
            p.get_current_measure().set_barline(style=bs)

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path

from musicxml.xmlelement.xmlelement import *

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase, create_technical, create_articulation, create_ornament
from musicscore.util import XML_TECHNICAL_CLASSES, XML_ARTICULATION_CLASSES, XML_OTHER_NOTATIONS, XML_DYNAMIC_CLASSES, XML_ORNAMENT_CLASSES


class TestHelloWorldTechnicals(IdTestCase):
    def test_export_hello_world(self):
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a list of chords with technicals to a part
        """
        p = s.add_child(Part('P1'))
        for index, cls in enumerate(XML_ARTICULATION_CLASSES):
            next_cls = XML_ARTICULATION_CLASSES[index + 1] if index != len(XML_ARTICULATION_CLASSES) - 1 else None
            ch = Chord(60, 1)
            ch.add_x(create_articulation(cls))
            p.add_chord(ch)
            if next_cls:
                ch = Chord(60, 1)
                ch.add_x(create_articulation(cls))
                ch.add_x(create_articulation(next_cls))
                p.add_chord(ch)

        for index, cls in enumerate(XML_TECHNICAL_CLASSES):
            if cls == XMLHammerOn:
                ch = Chord(60, 1)
                ch.add_x(XMLHammerOn('2', type='start'))
                p.add_chord(ch)
                ch = Chord(60, 1)
                ch.add_x(XMLHammerOn('2', type='stop'))
            elif cls == XMLPullOff:
                ch = Chord(60, 1)
                ch.add_x(XMLPullOff('2', type='start'))
                p.add_chord(ch)
                ch = Chord(60, 1)
                ch.add_x(XMLPullOff('2', type='stop'))
            else:
                next_cls = XML_TECHNICAL_CLASSES[index + 1] if index != len(XML_TECHNICAL_CLASSES) - 1 else None
                ch = Chord(60, 1)
                ch.add_x(create_technical(cls))
                p.add_chord(ch)
                if next_cls and next_cls not in [XMLHammerOn, XMLPullOff]:
                    ch = Chord(60, 1)
                    ch.add_x(create_technical(cls))
                    ch.add_x(create_technical(next_cls))
                    ch.add_x(XMLAccent(placement='above'))
                    p.add_chord(ch)

        for cls in XML_ORNAMENT_CLASSES[1:]:
            ch = Chord(60, 1)
            ch.add_x(create_ornament(cls))
            p.add_chord(ch)
        ch = Chord(60, 4)
        p.add_chord(ch)
        ch.add_x(XMLTrillMark())
        ch.add_x(XMLAccidentalMark('sharp'))
        ch.add_x(XMLWavyLine(type='start'))
        ch.add_x(XMLWavyLine(type='stop'))

        for cls in XML_DYNAMIC_CLASSES:
            ch = Chord(60, 1)
            ch.add_x(cls(), placement='below')
            p.add_chord(ch)

        for cls in XML_OTHER_NOTATIONS:
            if cls in [XMLGlissando, XMLSlide, XMLSlur]:
                ch = Chord(60, 1)
                ch.add_x(cls(type='start'))
                p.add_chord(ch)
                ch = Chord(60, 1)
                ch.add_x(cls(type='stop'))
                p.add_chord(ch)
            elif cls == XMLNonArpeggiate:
                ch = Chord(60, 1)
                ch.add_x(cls(type='top'))
                p.add_chord(ch)
            elif cls == XMLOtherNotation:
                ch = Chord(60, 1)
                ch.add_x(cls(type='single'))
                p.add_chord(ch)
            else:
                ch = Chord(60, 1)
                ch.add_x(cls())
                p.add_chord(ch)
        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



import inspect
import itertools
import xml.etree.ElementTree as ET
from difflib import Differ
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock, patch

import xmltodict
from deepdiff import DeepDiff
from quicktions import Fraction

from musicscore import generate_measures, Chord, C
from musicscore.exceptions import LyricsExtensionError, LyricSyllabicOrExtensionError
from musicscore.part import Id
from musicxml import XMLLyric, XMLExtend
from musicxml.xmlelement.xmlelement import *
from musicscore.util import XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS, XML_DIRECTION_TYPE_CLASSES, XML_OTHER_NOTATIONS, \
    XML_ORNAMENT_CLASSES, XML_ARTICULATION_CLASSES, XML_TECHNICAL_CLASSES, XML_DYNAMIC_CLASSES, \
    XML_ORNAMENT_AND_OTHER_NOTATIONS


class XMLsDifferException(Exception):
    pass


notehead_values = ['slash', 'triangle', 'diamond', 'square', 'rectangle', 'cross', 'x', 'circle dot', 'circle-x',
                   'circled', 'inverted triangle', 'left triangle', 'arrow down', 'arrow up', 'slashed',
                   'back slashed', 'normal', 'cluster', 'none']
notehead_aikin_values = ['do', 're', 'mi', 'fa', 'fa up', 'so', 'la', 'ti']


def check_notes(notes, midi_values, quarter_durations):
    if len(notes) != len(midi_values) or len(notes) != len(quarter_durations):
        raise ValueError
    for n, m, qd in zip(notes, midi_values, quarter_durations):
        assert n.midi.value == m
        assert n.quarter_duration == qd


class IdTestCase(TestCase):
    def setUp(self):
        Id.__refs__.clear()


class ChordTestCase(IdTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.mock_beat = Mock()
        self.mock_voice = Mock()
        self.mock_staff = Mock()
        self.mock_measure = Mock()

        self.mock_voice.number = 1
        self.mock_beat.up = self.mock_voice
        self.mock_voice.up = self.mock_staff
        self.mock_staff.up = self.mock_measure
        self.mock_measure.get_divisions.return_value = 1
        self.mock_staff.number = None

    def tearDown(self) -> None:
        patch.stopall()


def _create_expected_path(path):
    return Path(path.parent / f"{path.stem}_expected{path.suffix}")


def diff_xml(path_1, path_2=None):
    if path_2 is None:
        path_2 = _create_expected_path(path_1)

    with open(path_1) as f1, open(path_2) as f2:
        f_1 = [i.strip() for i in f1]
        f_2 = [i.strip() for i in f2]

    diff = Differ()
    difference = list(diff.compare(f_1, f_2))
    return [d for d in difference if d.startswith('-') or d.startswith('+')]


def generate_all_16ths():
    output = [tuple(4 * [Fraction(1, 4)])]
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 4), Fraction(1, 4), Fraction(1, 2)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 4), Fraction(3, 4)]))))
    return output


def generate_all_32nds():
    output = [tuple(8 * [Fraction(1, 8)])]
    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(2, 8)] + 6 * [Fraction(1, 8)]))))
    output.extend(list(dict.fromkeys(itertools.permutations(2 * [Fraction(2, 8)] + 4 * [Fraction(1, 8)]))))
    output.extend(list(dict.fromkeys(itertools.permutations(3 * [Fraction(2, 8)] + 2 * [Fraction(1, 8)]))))

    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(3, 8)] + 5 * [Fraction(1, 8)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations(1 * [Fraction(3, 8)] + 1 * [Fraction(2, 8)] + 3 * [Fraction(1, 8)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations(1 * [Fraction(3, 8)] + 2 * [Fraction(2, 8)] + 1 * [Fraction(1, 8)]))))
    output.extend(list(dict.fromkeys(itertools.permutations(2 * [Fraction(3, 8)] + 2 * [Fraction(1, 8)]))))
    output.extend(list(dict.fromkeys(itertools.permutations(2 * [Fraction(3, 8)] + 1 * [Fraction(2, 8)]))))

    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(4, 8)] + 4 * [Fraction(1, 8)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations(1 * [Fraction(4, 8)] + 1 * [Fraction(2, 8)] + 2 * [Fraction(1, 8)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations(1 * [Fraction(4, 8)] + 1 * [Fraction(3, 8)] + 1 * [Fraction(1, 8)]))))

    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(5, 8)] + 3 * [Fraction(1, 8)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations(1 * [Fraction(5, 8)] + 1 * [Fraction(2, 8)] + 1 * [Fraction(1, 8)]))))
    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(5, 8)] + 1 * [Fraction(3, 8)]))))

    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(6, 8)] + 2 * [Fraction(1, 8)]))))

    output.extend(list(dict.fromkeys(itertools.permutations(1 * [Fraction(7, 8)] + 1 * [Fraction(1, 8)]))))
    return output


def generate_all_quintuplets():
    output = [tuple(5 * [Fraction(1, 5)])]
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 5), Fraction(1, 5), Fraction(1, 5), Fraction(2, 5)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 5), Fraction(2, 5), Fraction(2, 5)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 5), Fraction(1, 5), Fraction(3, 5)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(2, 5), Fraction(3, 5)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 5), Fraction(4, 5)]))))
    return output


def generate_all_sextuplets():
    output = [tuple(6 * [Fraction(1, 6)])]
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(1, 6),
                                                   Fraction(2, 6)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(1, 6), Fraction(2, 6),
                                                             Fraction(2, 6)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(1, 6), Fraction(1, 6), Fraction(3, 6)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(2, 6), Fraction(3, 6)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(1, 6), Fraction(4, 6)]))))
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 6), Fraction(5, 6)]))))
    return output


def generate_all_septuplets():
    output = [tuple(7 * [Fraction(1, 7)])]
    output.extend(list(dict.fromkeys(
        itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(1, 7),
                                Fraction(2, 7)]))))
    output.extend(
        list(dict.fromkeys(
            itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(2, 7), Fraction(2, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(2, 7), Fraction(2, 7), Fraction(2, 7)]))))
    output.extend(
        list(dict.fromkeys(
            itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(3, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(2, 7), Fraction(3, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(3, 7), Fraction(3, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(1, 7), Fraction(4, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(2, 7), Fraction(4, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(3, 7), Fraction(4, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(1, 7), Fraction(5, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(2, 7), Fraction(5, 7)]))))
    output.extend(
        list(dict.fromkeys(itertools.permutations([Fraction(1, 7), Fraction(6, 7)]))))
    return output


def generate_all_triplets():
    output = [tuple(3 * [Fraction(1, 3)])]
    output.extend(list(dict.fromkeys(itertools.permutations([Fraction(1, 3), Fraction(2, 3)]))))
    return output


def get_xml_elements_diff(el1, el2):
    return DeepDiff(xmltodict.parse(ET.tostring(el1)), xmltodict.parse(ET.tostring(el2)))


def get_xml_diff_part(expected, path, file_path):
    el1 = ET.parse(file_path.parent / path).getroot().find("part[@id='part-1']")
    el2 = ET.parse(file_path.parent / expected).getroot().find("part[@id='part-1']")
    diff = get_xml_elements_diff(el1=el1, el2=el2)
    if diff:
        raise XMLsDifferException(diff)


def generate_xml_file(score, *simpleformats, path):
    part = score.add_part(id='part-1')
    for index, simpleformat in enumerate(simpleformats):
        for chord in simpleformat.chords:
            part.add_chord(chord, staff_number=index + 1)
    score.export_xml(path)


def generate_repetitions(part):
    for m in generate_measures([(3, 8), (3, 4)] + 2 * [(6, 4)] + 3 * [(4, 4)] + [(5, 4)]):
        part.add_child(m)

    for qd in 3 * [0.5] + 3 * [1] + 4 * [1.5] + 3 * [2] + [1, 1, 1, 2, 1, 1, 2, 2, 3, 3, 3, 4, 4, 2, 2, 1.5, 1.5,
                                                           3.25, 1.75, 1]:
        part.add_chord(Chord(C(5, '#'), qd))


def generate_path(frame):
    f = str(inspect.getframeinfo(frame).function) + '.xml'
    path = Path(inspect.getframeinfo(frame).filename).parent / f
    return path


def create_test_objects(type_):
    output = []
    if type_ == 'direction_type':
        for cl in XML_DIRECTION_TYPE_CLASSES + XML_DIRECTION_TYPE_AND_OTHER_NOTATIONS:
            if cl == XMLSymbol:
                obj = cl('0')
            elif cl == XMLWedge:
                obj = cl(type='crescendo')
            elif cl == XMLDashes:
                obj = cl(type='start')
            elif cl == XMLBracket:
                obj = cl(type='start', line_end='none')
            elif cl == XMLPedal:
                obj = cl(type='start')
            elif cl == XMLMetronome:
                obj = cl()
                obj.add_child(XMLBeatUnit('quarter'))
                obj.add_child(XMLPerMinute('120'))
            elif cl == XMLOctaveShift:
                obj = cl(type='up')
            elif cl == XMLHarpPedals:
                obj = cl()
                pt = obj.add_child(XMLPedalTuning())
                pt.add_child(XMLPedalStep('A'))
                pt.add_child(XMLPedalAlter(1))
            elif cl == XMLStringMute:
                obj = cl(type='on')
            elif cl == XMLScordatura:
                obj = cl()
                acc = obj.add_child(XMLAccord())
                acc.add_child(XMLTuningStep('A'))
                acc.add_child(XMLTuningOctave(0))
            # elif cl == XMLImage:
            #     obj = cl(source='www.example.com', type='image/gif')
            elif cl == XMLPrincipalVoice:
                obj = cl(type='start', symbol='none')
            elif cl == XMLPercussion:
                obj = cl()
                obj.add_child(XMLWood('cabasa'))
            elif cl == XMLStaffDivide:
                obj = cl(type='up')
            else:
                obj = cl()
            output.append(obj)

    elif type_ == 'technical':
        needed_values = {XMLFret: 1, XMLString: 1, XMLHandbell: 'belltree', XMLFingering: '2', XMLPluck: 'something',
                         XMLTap: '2', XMLHarmonClosed: 'yes', XMLOtherTechnical: 'something'}
        needed_types = {XMLHammerOn: 'start', XMLPullOff: 'start'}
        needed_children = {XMLBend: XMLBendAlter(2), XMLHole: XMLHoleClosed('yes'), XMLArrow: XMLArrowDirection('up'),
                           XMLHarmonMute: XMLHarmonClosed('yes')}
        for cl in XML_TECHNICAL_CLASSES:
            if cl in needed_children:
                obj = cl()
                obj.add_child(needed_children[cl])
            elif cl in needed_types:
                obj = cl(type=needed_types[cl])
            elif cl in needed_values:
                obj = cl(needed_values[cl])
            else:
                obj = cl()
            output.append(obj)

    elif type_ == 'ornament':
        for cl in XML_ORNAMENT_AND_OTHER_NOTATIONS + XML_ORNAMENT_CLASSES:
            if cl == XMLAccidentalMark:
                obj = cl('sharp')
            elif cl == XMLTremolo:
                obj = cl(3)
            else:
                obj = cl()
            output.append(obj)

    elif type_ == 'notation':
        for cl in XML_OTHER_NOTATIONS + XML_ORNAMENT_AND_OTHER_NOTATIONS:
            if cl == XMLAccidentalMark:
                obj = cl('sharp')
            else:
                obj = cl()
            output.append(obj)

    elif type_ == 'articulation':
        for cl in XML_ARTICULATION_CLASSES:
            obj = cl()
            output.append(obj)
    elif type_ == 'dynamics':
        for cl in XML_DYNAMIC_CLASSES:
            obj = cl()
            output.append(obj)
    else:
        raise NotImplementedError(f"type_ {type_}")
    return output


def _generate_xml_lyric(text=None, number=1, syllabic=None, extend=None, **kwargs):
    xl = XMLLyric(number=str(number), **kwargs)
    xl.xml_text = text
    xl.xml_syllabic = syllabic
    if extend:
        xl.xml_extend = XMLExtend(type=extend)
    return xl



from pathlib import Path

from musicscore import Score, Chord
from musicscore.tests.util import IdTestCase


class TestParts(IdTestCase):
    def test_multiple_parts(self):
        """
        Create multiple parts with names
        """
        score = Score()
        for i in range(5):
            p = score.add_part(f'part-{i}')
            p.add_chord(Chord(0, 4))

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from pathlib import Path

from musicscore import Score, G, A, B, C, D, E, F, Chord, Midi, Accidental, Time
from musicscore.tests.util import IdTestCase

"""
All pitch intervals in ascending jump size.
"""


class TestLily01b(IdTestCase):
    def test_lily_01b_Pitches_Intervals(self):
        score = Score('Various pitches and interval sizes')
        p = score.add_part('intervals')
        p.add_measure(Time(2, 4))
        for i in range(24):
            m1 = Midi(72 + i)
            m2 = Midi(72 - i)
            p.add_chord(Chord(m1, 1))
            p.add_chord(Chord(m2, 1))

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from typing import List, Union, Optional

from musicscore import Part, Chord
from musicscore.chord import Rest
from musicscore.exceptions import AlreadyFinalizedError, ScoreMultiMeasureRestError
from musicscore.finalize import FinalizeMixin
from musicscore.quantize import QuantizeMixin
from musicxml.xmlelement.xmlelement import XMLScorePartwise, XMLPartList, XMLCredit, XMLCreditWords, XMLIdentification, \
    XMLEncoding, \
    XMLSupports, XMLScorePart, XMLPartGroup, XMLGroupSymbol, XMLGroupBarline, XMLGroupName, XMLGroupAbbreviation, \
    XMLMeasureStyle

from musicscore.musictree import MusicTree
from musicscore.quarterduration import QuarterDuration
from musicscore.xmlwrapper import XMLWrapper
from musicscore.layout import Scaling, PageLayout, SystemLayout, StaffLayout

__all__ = ['TITLE', 'SUBTITLE', 'POSSIBLE_SUBDIVISIONS', 'Score']
#:
TITLE = {'font_size': 24, 'default_x': {'A4': {'portrait': 616}}, 'default_y': {'A4': {'portrait': 1573}},
         'justify': 'center',
         'valign': 'top'}

#:
SUBTITLE = {'font_size': 18, 'default_x': {'A4': {'portrait': 616}}, 'default_y': {'A4': {'portrait': 1508}},
            'halign': 'center',
            'valign': 'top'}
#:
POSSIBLE_SUBDIVISIONS = {QuarterDuration(1, 4): [2, 3], QuarterDuration(1, 2): [2, 3, 4, 5],
                         QuarterDuration(1): [2, 3, 4, 5, 6, 7, 8]}


class Score(MusicTree, QuantizeMixin, FinalizeMixin, XMLWrapper):
    _ATTRIBUTES = {'version', 'title', 'subtitle', 'scaling', 'page_layout', 'system_layout', 'staff_layout'}
    _ATTRIBUTES = _ATTRIBUTES.union(MusicTree._ATTRIBUTES)
    _ATTRIBUTES = _ATTRIBUTES.union(QuantizeMixin._ATTRIBUTES)

    XMLClass = XMLScorePartwise

    def __init__(self, version='4.0', title=None, subtitle=None, get_quantized=False, *args, **kwargs):

        super().__init__(get_quantized=get_quantized)
        self._xml_object = self.XMLClass(*args, **kwargs)
        self._update_xml_object()
        self._version = None
        self._title = None
        self._subtitle = None
        self._page_layout = None
        self._system_layout = None
        self._staff_layout = None
        self._scaling = None

        self.scaling = Scaling()
        self.page_layout = PageLayout()
        self.system_layout = SystemLayout()
        self.version = version
        self.title = title
        self.subtitle = subtitle
        self._possible_subdivisions = POSSIBLE_SUBDIVISIONS.copy()

        self._measure_numbers_within_multi_measure_rests = set()

        self._final_updated = False

    def _create_missing_measures(self):
        number_of_measures = max([len(p.get_children()) for p in self.get_children()])
        longest_parts = [p for p in self.get_children() if len(p.get_children()) == number_of_measures]
        for p in set(self.get_children()).difference(set(longest_parts)):
            for measure_number in range(len(p.get_children()) + 1, number_of_measures + 1):
                p.add_chord(Rest(longest_parts[0].get_measure(measure_number).quarter_duration))

    def _get_title_attributes(self):
        output = TITLE.copy()
        output['default_x'] = TITLE['default_x']['A4']['portrait']
        output['default_y'] = TITLE['default_y']['A4']['portrait']
        return output

    def _get_subtitle_attributes(self):
        output = SUBTITLE.copy()
        output['default_x'] = SUBTITLE['default_x']['A4']['portrait']
        output['default_y'] = SUBTITLE['default_y']['A4']['portrait']
        return output

    def _set_last_barline(self):
        last_measures = [p.get_children()[-1] for p in self.get_children() if p.get_children()]
        try:
            if not last_measures[0].get_barline():
                for m in last_measures:
                    m.set_barline(style='light-heavy')
        except IndexError:
            pass

    def _set_missing_barlines(self):
        for measures in zip(*[p.get_children() for p in self.get_children()]):
            right_barline = None
            left_barline = None
            for m in measures:
                if m.get_barline(location='right'):
                    right_barline = m.get_barline(location='right')
                    break
            for m in measures:
                if m.get_barline(location='left'):
                    left_barline = m.get_barline(location='left')
                    break
            if right_barline:
                for m in measures:
                    m._barlines['right'] = right_barline
            if left_barline:
                for m in measures:
                    m._barlines['left'] = left_barline

    def _update_xml_object(self):
        self.xml_object.xml_part_list = XMLPartList()
        self.xml_object.xml_identification = XMLIdentification()
        encoding = self.xml_object.xml_identification.xml_encoding = XMLEncoding()
        encoding.add_child(XMLSupports(attribute='new-system', element='print', type='yes', value='yes'))
        encoding.add_child(XMLSupports(attribute='new-page', element='print', type='yes', value='yes'))

        encoding.add_child(XMLSupports(element='accidental', type='yes'))
        encoding.add_child(XMLSupports(element='beam', type='yes'))
        encoding.add_child(XMLSupports(element='stem', type='yes'))

    @property
    def page_layout(self) -> PageLayout:
        """
        Sets and gets page layout. After setting value, page layout's parent is set to self.

        :type: :obj:`~musicscore.pagelayout.PageLayout`
        :return: :obj:`~musicscore.pagelayout.PageLayout`
        """
        return self._page_layout

    @page_layout.setter
    def page_layout(self, val):
        if not isinstance(val, PageLayout):
            raise TypeError
        self._page_layout = val
        self.page_layout.parent = self

    @property
    def scaling(self) -> Scaling:
        """
        :type: :obj:`~musicscore.scaling.Scaling`
        :return: :obj:`~musicscore.scaling.Scaling`
        """
        return self._scaling

    @scaling.setter
    def scaling(self, val):
        if not isinstance(val, Scaling):
            raise TypeError
        val.score = self
        self._scaling = val

    @property
    def staff_layout(self):
        """
        Sets and gets staff layout. After setting value, staff layout's parent is set to self.

        :type: :obj:`~musicscore.stafflayout.StaffLayout`
        :return: :obj:`~musicscore.stafflayout.StaffLayout`
        """

        return self._staff_layout

    @staff_layout.setter
    def staff_layout(self, val):
        if not isinstance(val, StaffLayout):
            raise TypeError
        self._staff_layout = val
        self.staff_layout.parent = self

    @property
    def subtitle(self):
        """
        - If val is ``None`` and a subtitle object as credit already exists, this object will be removed.
        - If val is not ``None``:
          - if a subtitle exists its ``value_`` will be replaced.
          - else a subtitle as credit will be added.

        :type: Optional[str]
        :return: Optional[str]
        """
        return self._subtitle

    @subtitle.setter
    def subtitle(self, val):
        if val is not None:
            if self._subtitle is None:
                credit = self.xml_object.add_child(XMLCredit(page=1))
                credit.xml_credit_type = 'subtitle'
                self._subtitle = credit.add_child(XMLCreditWords(value_=val, **self._get_subtitle_attributes()))
            else:
                self._subtitle.value_ = val
        else:
            if self._subtitle is None:
                pass
            else:
                credit = self._subtitle.up
                credit.up.remove(credit)
                self._subtitle = None

    @property
    def system_layout(self):
        """
        Sets and gets system layout. After setting value, system layout's parent is set to self.

        :type: :obj:`~musicscore.systemlayout.SystemLayout`
        :return: :obj:`~musicscore.systemlayout.SystemLayout`
        """
        return self._system_layout

    @system_layout.setter
    def system_layout(self, val):
        if not isinstance(val, SystemLayout):
            raise TypeError
        self._system_layout = val
        self.system_layout.parent = self

    @property
    def title(self):
        """
        - If val is ``None`` and a title object as credit already exists, this object will be removed.
        - If val is not ``None``:
          - if a title exists its ``value_`` will be replaced.
          - else a title as credit will be added.

        :type: Optional[str]
        :return: Optional[str]
        """
        return self._title

    @title.setter
    def title(self, val):
        if val is not None:
            if self._title is None:
                credit = self.xml_object.add_child(XMLCredit(page=1))
                credit.xml_credit_type = 'title'
                self._title = credit.add_child(XMLCreditWords(value_=val, **self._get_title_attributes()))
            else:
                self._title.value_ = val
        else:
            if self._title is None:
                pass
            else:
                credit = self._title.up
                credit.up.remove(credit)
                self._title = None

    @property
    def version(self) -> str:
        """
        :type: Any
        :return: xml_object's version
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, val):
        self._version = str(val)
        self.xml_object.version = self.version

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def add_child(self, child: 'Part') -> 'Part':
        """
        - Check and add child to list of children. Child's parent is set to self.
        - Part's ``xml_object`` is add to score's ``xml_object`` as child
        - Part's ``score_part.xml_object`` is added to score's ``xml_part_list``

        :param child: :obj:`~musicscore.part.Part`, required
        :return: child
        :rtype: :obj:`~musicscore.part.Part`
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_child')
        super().add_child(child)
        self.xml_object.add_child(child.xml_object)
        if not self.xml_part_list:
            self.xml_part_list.xml_score_part = child.score_part.xml_object
        else:
            self.xml_part_list.add_child(child.score_part.xml_object)
        return child

    def add_part(self, id: str) -> 'Part':
        """
        Creates and adds part

        :param id: part's id
        :return: part
        """
        if self._finalized is True:
            raise AlreadyFinalizedError(self, 'add_part')
        p = Part(id)
        return self.add_child(p)

    def export_xml(self, path: 'pathlib.Path') -> None:
        """
        Creates a musicxml file

        :param path: Output xml file
        :return: None
        """
        with open(path, '+w') as f:
            f.write("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE score-partwise PUBLIC
    "-//Recordare//DTD MusicXML 4.0 Partwise//EN"
    "http://www.musicxml.org/dtds/partwise.dtd">
""")
            f.write(self.to_string())

    def finalize(self) -> None:
        self._create_missing_measures()
        self._set_missing_barlines()
        self._set_last_barline()
        super().finalize()
        for measure_number in self._measure_numbers_within_multi_measure_rests:
            for part in self.get_children():
                measure = part.get_measure(measure_number)
                for ch in measure.get_chords():
                    ch.notes[0].xml_rest.measure = 'yes'

    def get_children(self) -> List['Part']:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.part.Part`]
        """
        return super().get_children()

    def group_parts(self, number: Union[int, str], start_part_number: int, end_part_number: int, symbol: str = 'square',
                    name: Optional[str] = None, abbreviation: Optional[str] = None,
                    **kwargs) -> None:
        """
        Adds a XMLPartList child

        :param number: sets number property of :obj:`~musicxml.xmlelement.xmlelement.XMLPartGroup`
        :param start_part_number: number of part to start the group
        :param end_part_number: number of part to end the group
        :param symbol: symbol property of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupSymbol`:'none', 'brace', 'line', 'bracket', 'square'; default: 'square'
        :param name: value of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupName`
        :param abbreviation: value of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupAbbreviation`
        :param kwargs: kwargs of :obj:`~musicxml.xmlelement.xmlelement.XMLGroupSymbol`
        :return: None
        """
        parts = self.get_children_of_type(Part)
        for part_number in [start_part_number, end_part_number]:
            if not 0 < part_number <= len(parts):
                raise ValueError(f'Wrong part number {part_number}. Permitted between 1 and {len(parts)}')
        if not start_part_number < end_part_number:
            raise ValueError(
                f'Wrong part numbers. start_part_number {start_part_number} must be smaller than end_part_number {end_part_number}')

        new_xml_part_list = XMLPartList(xsd_check=False)
        for child in self.xml_part_list.get_children():
            if isinstance(child, XMLScorePart) and child.id == parts[start_part_number - 1].id_.value:
                pg = new_xml_part_list.add_child(XMLPartGroup(number=str(number), type='start'))
                pg.add_child(XMLGroupSymbol(symbol, **kwargs))
                pg.add_child(XMLGroupBarline('yes'))
                if name:
                    pg.add_child(XMLGroupName(name))
                if abbreviation:
                    pg.add_child(XMLGroupAbbreviation(abbreviation))
            new_xml_part_list.add_child(child)
            if isinstance(child, XMLScorePart) and child.id == parts[end_part_number - 1].id_.value:
                new_xml_part_list.add_child(XMLPartGroup(number=str(number), type='stop'))

        self.xml_part_list = new_xml_part_list

    def set_multi_measure_rest(self, first_measure_number: int, last_measure_number: int) -> None:
        """
        Creates a multi measure rest

        :param first_measure_number: number of measure to start the multi measure rest
        :param last_measure_number: number of measure to end the multi measure rest
        :return: None
        """

        if len(self.get_children()) == 0:
            raise ScoreMultiMeasureRestError(f'score has no parts.')
        if last_measure_number <= first_measure_number:
            raise ScoreMultiMeasureRestError(
                f'last_measure_number {last_measure_number} must be larger than first_measure_number {first_measure_number}')
        for x in range(first_measure_number, last_measure_number + 1):
            if x in self._measure_numbers_within_multi_measure_rests:
                raise ScoreMultiMeasureRestError(f'measure number {x} is already part of a multiple measure reset')
        for part in self.get_children():
            for _ in range(last_measure_number - len(part.get_children())):
                m = part.add_measure()
                part.add_chord(Chord(0, m.quarter_duration))

            for measure in part.get_children()[first_measure_number - 1:last_measure_number]:
                for ch in measure.get_chords():
                    if not ch.is_rest:
                        raise ScoreMultiMeasureRestError(
                            f'Measures contain not rest chords')

            first_measure = part.get_measure(first_measure_number)
            if not first_measure.xml_attributes.xml_measure_style:
                first_measure.xml_attributes.xml_measure_style = XMLMeasureStyle()
            first_measure.xml_attributes.xml_measure_style.xml_multiple_rest = last_measure_number - first_measure_number + 1
            if part == self.get_children()[0]:
                self._measure_numbers_within_multi_measure_rests.update(
                    {x for x in range(first_measure_number, last_measure_number + 1)})

    def write(self, *args, **kwargs):
        """
        Not implemented. Use Score.export_xml instead!
        """
        raise NotImplementedError('Use Score.export_xml instead!')



from pathlib import Path

from musicxml.xmlelement.xmlelement import XMLLyric

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloLyrics(IdTestCase):
    def test_export_hello_world_lyrics(self):
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a part
        """
        p = s.add_child(Part('P1', name='Music'))
        """
        He adds some chords with lyrics
        """
        for i in range(6):
            ch = Chord(60 + i, 3)
            ch.add_lyric(ch.midis[0].accidental.sign)
            p.add_chord(ch)
        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



"""Several spanners defined in MusicXML: tuplet, slur (solid, dashed), tie, wedge (cresc, dim), tr + wavy-line,
single-note trill spanner, octave-shift (8va,15mb), bracket (solid down/down, dashed down/down, solid none/down,
dashed none/up, solid none/none), dashes, glissando (wavy), bend-alter, slide (solid), grouping, two-note tremolo,
hammer-on, pull-off, pedal (down, change, up)."""
from pathlib import Path

from musicscore import Score, Time, Chord, F, C
from musicscore.tests.util import IdTestCase
from musicscore.util import slur_chords, wedge_chords, trill_chords, bracket_chords, octave_chords
from musicxml import XMLDashes, XMLGlissando, XMLBendAlter, XMLBend, XMLSlide, XMLHammerOn, XMLPullOff, XMLPedal


class TestLily33a(IdTestCase):
    def test_lily_33a_Spanners(self):
        score = Score()
        part = score.add_part('p1')

        t = Time(3, 4)
        t.actual_signatures = [1, 2, 1, 4]
        part.add_measure(t)
        part.add_measure(Time(3, 4))

        [part.add_chord(Chord(71, 2 / 3)) for _ in range(3)]
        part.add_chord(Chord(0, 1))

        chords = [Chord(71, 1) for _ in range(3)]
        slur_chords(chords)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        slur_chords(chords, line_type='dashed')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        wedge_chords(chords, 'crescendo', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        wedge_chords(chords, 'diminuendo', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        trill_chords(chords, placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(0, 2)]
        trill_chords(chords, placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        octave_chords(chords)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        octave_chords(chords, type='up', size=15)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, 'solid', 'down', 'down', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, 'dashed', 'down', 'down', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, 'solid', 'none', 'down', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, 'dashed', 'none', 'up', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        bracket_chords(chords, 'solid', 'none', 'none', placement='above')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        chords[0].add_words('rit.', placement='above', font_style='italic')
        chords[0].add_x(XMLDashes(type='start', relative_x=10))
        chords[1].add_x(XMLDashes(type='continue'))
        chords[2].add_x(XMLDashes(type='stop'))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(F(5), 1), Chord(0, 1)]
        chords[0].add_x(XMLGlissando(type='start', line_type='wavy'))
        chords[1].add_x(XMLGlissando(type='stop'))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(F(5), 1), Chord(0, 1)]
        b = chords[0].add_x(XMLBend(shape='curved'))
        b.xml_bend_alter = XMLBendAlter(5)
        b = chords[1].add_x(XMLBend())
        b.xml_bend_alter = XMLBendAlter(0)
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(C(4), 1), Chord(0, 1)]
        chords[0].add_x(XMLSlide(type='start'))
        chords[1].add_x(XMLSlide(type='stop'))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        chords[0].add_words('XMLGrouping not implemented yet')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(71, 1), Chord(0, 1)]
        chords[0].add_words('Fingered tremolo not implemented yet', placement='below')
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(71, 1), Chord(0, 1)]
        chords[0].add_x(XMLHammerOn(type='start'))
        chords[1].add_x(XMLHammerOn(type='stop'))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1), Chord(71, 1), Chord(0, 1)]
        chords[0].add_x(XMLPullOff(type='start'))
        chords[1].add_x(XMLPullOff(type='stop'))
        [part.add_chord(ch) for ch in chords]

        chords = [Chord(71, 1) for _ in range(3)]
        chords[0].add_x(XMLPedal(type='start'))
        chords[1].add_x(XMLPedal(type='change'))
        chords[2].add_x(XMLPedal(type='stop'))
        [part.add_chord(ch) for ch in chords]

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



class XMLWrapper:
    """
    XMLWrapper contains an xml object at its core. It is the place where all the intuitive stuff is translated to attributes and children of
    xml objects. All attributes and properties which are not listed in _ATTRIBUTES will be set to or get from core xml object.
    """
    _ATTRIBUTES = {}

    XMLClass = None

    @property
    def xml_object(self) -> 'XMLELement':
        """
        :return: wrapped musicxml element of type :obj:`XMLClass`
        """
        return self._xml_object

    def to_string(self, *args, **kwargs) -> str:
        """
        Calls :obj:`~musicxml.xmlelement.xmlelement.XMLElement.to_string` method of self.xml_object

        :return: musicxml snippet
        """
        if self.xml_object:
            return self.xml_object.to_string(*args, **kwargs)
        else:
            raise ValueError(f'{self.__class__.__name__} has no xml object.')

    def __setattr__(self, key, value):
        attributes = self._ATTRIBUTES
        try:
            if self._TREE_ATTRIBUTES:
                attributes.union(self._TREE_ATTRIBUTES)
        except AttributeError:
            pass
        if key[0] != '_' and ('_xml_object' in self.__dict__ and key not in attributes and key not in self.__dict__):
            setattr(self._xml_object, key, value)
        else:
            super().__setattr__(key, value)

    def __getattr__(self, item):
        if item == '_TREE_ATTRIBUTES':
            raise AttributeError
        if item == 'xml_object':
            return super().__getattribute__(item)
        try:
            return self._xml_object.__getattribute__(item)
        except AttributeError:
            try:
                return self._xml_object.__getattr__(item)
            except AttributeError:
                return super().__getattribute__(item)



from unittest import TestCase

from musicscore import Part
from musicscore.clef import Clef, TrebleClef, BassClef, TenorClef, AltoClef, PercussionClef
from musicscore.tests.util import IdTestCase


class TestClef(IdTestCase):
    def test_default_clef(self):
        c = Clef()
        expected = """<clef>
  <sign>G</sign>
  <line>2</line>
</clef>
"""
        assert c.to_string() == expected

    def test_clef_copy(self):
        c = Clef('F', 2, show=False, octave_change=-1)
        copied = c.__copy__()
        assert copied != c
        assert copied.xml_object != c.xml_object
        assert copied.sign == c.sign
        assert copied.line == c.line
        assert copied.show == c.show
        assert copied.octave_change == c.octave_change

    def test_global_clefs(self):
        c = TrebleClef()
        assert c.sign == 'G'
        assert c.line == 2
        c = BassClef()
        assert c.sign == 'F'
        assert c.line == 4
        c = TenorClef()
        assert c.sign == 'C'
        assert c.line == 4
        c = AltoClef()
        assert c.sign == 'C'
        assert c.line == 3
        c = PercussionClef()
        assert c.sign == 'percussion'
        assert c.line is None

        c = TrebleClef(octave_change=1)
        expected = """<clef>
  <sign>G</sign>
  <line>2</line>
  <clef-octave-change>1</clef-octave-change>
</clef>
"""
        assert c.to_string() == expected

    def test_staff_clef(self):
        clefs = [None, TrebleClef(), AltoClef(), TenorClef(), BassClef(), PercussionClef(),
                 TrebleClef(octave_change=-1),
                 Clef(sign='F', line=3)]

        part = Part('p1')
        for c in clefs:
            m = part.add_measure()
            m.get_staff(1).clef = c
        for m, c in zip(part.get_children(), clefs):
            if c:
                assert m.clefs == [c]
        part.finalize()



import random
from pathlib import Path

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase


class TestHelloCoplexAccidentals(IdTestCase):
    def test_export_hello_world_complex_accidentals(self):
        """
        Tester creates a timewise score
        """
        s = Score()
        """
        He adds a part
        """
        p = s.add_child(Part('P1', name='Music'))
        """
        He adds a lot of random pitches with different quarter_durations
        """
        random.seed(10)
        midis = [random.randint(60, 72) for _ in range(200)]
        quarter_durations = [random.randint(1, 24) / 4 for _ in range(200)]
        for m, qd in zip(midis, quarter_durations):
            ch = Chord(m, qd)
            ch.add_lyric(ch.midis[0].accidental.sign)
            p.add_chord(ch)

        """
        ... and exports the xml
        """
        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)



import copy
from math import log2
from typing import Optional, Union, List

from musicscore import STANDARD, ENHARMONIC, FLAT, SHARP, FORCEFLAT, FORCESHARP
from musicscore.exceptions import AlreadyFinalizedError
from musicxml.xmlelement.xmlelement import *  # type: ignore
from musicxml.xsd.xsdsimpletype import XSDSimpleTypeNoteheadValue  # type: ignore

from musicscore.accidental import Accidental
from musicscore.musictree import MusicTree

__all__ = ['Midi', 'MidiNote', 'C', 'D', 'E', 'F', 'G', 'A', 'B', 'midi_to_frequency', 'frequency_to_midi',
           'get_accidental_mode']


def midi_to_frequency(midi, a4=440):
    try:
        midi = midi.value
    except AttributeError:
        pass

    f = 2 ** ((midi - 69) / 12) * a4
    return f


def frequency_to_midi(frequency, a4=440):
    m = 69 + 12 * log2(frequency / a4)
    return m


def get_accidental_mode(midi_value: Union[float, int], accidental_sign: Optional[str] = None) -> str:
    """'
    :param midi_value: a valid midi value in half steps (int or float)
    :param accidental_sign: 'double-flat', 'flat-flat', 'bb', 'ff' – 'three-quarters-flat' – 'flat', 'b', 'f' – 'quarter-flat' – None, 'natural' – 'quarter-sharp' – 'sharp', '#', 's' – 'three-quarters-sharp' – 'double-sharp', 'sharp-sharp', 'x', '##', 'ss'
    :return: accidental_mode: 'standard', 'enharmonic', 'flat', 'sharp', 'force-flat', 'force-sharp'
    '"""
    # midi value is valid
    Midi(midi_value)
    if accidental_sign in [None, 'natural']:
        accidental_value = 0
    elif accidental_sign in ['quarter-sharp']:
        accidental_value = 0.5
    elif accidental_sign in ['sharp', '#', 's']:
        accidental_value = 1
    elif accidental_sign in ['three-quarters-sharp']:
        accidental_value = 1.5
    elif accidental_sign in ['double-sharp', 'sharp-sharp', 'x', '##', 'ss']:
        accidental_value = 2
    elif accidental_sign in ['quarter-flat']:
        accidental_value = -0.5
    elif accidental_sign in ['flat', 'b', 'f']:
        accidental_value = -1
    elif accidental_sign in ['three-quarters-flat']:
        accidental_value = -1.5
    elif accidental_sign in ['double-flat', 'flat-flat', 'bb', 'ff']:
        accidental_value = -2
    else:
        raise NotImplementedError(accidental_sign)

    if STANDARD[midi_value % 12][1] == accidental_value:
        return 'standard'
    elif ENHARMONIC[midi_value % 12][1] == accidental_value:
        return 'enharmonic'
    elif FLAT[midi_value % 12][1] == accidental_value:
        return 'flat'
    elif SHARP[midi_value % 12][1] == accidental_value:
        return 'sharp'
    elif FORCEFLAT[midi_value % 12][1] == accidental_value:
        return 'force-flat'
    elif FORCESHARP[midi_value % 12][1] == accidental_value:
        return 'force-sharp'


class Midi(MusicTree):
    """
    Midi is the representation of a Pitch with its midi value, and accidental sign. This object is used to create a Chord
    consisting of one or more pitches. The midi representation of a rest is a Midi object with value 0.
    """

    def __init__(self, value: Union[float, int], accidental: Optional[Accidental] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = None
        self._accidental = None
        self._notehead = None
        self._pitch_or_rest = None
        self._parent_chord = None
        self._parent_note = None
        self._ties = set()

        self._staff_number = None

        self.value = value
        self.accidental = accidental

    def _set_parent_chord(self, value):
        if value is not None and 'Chord' not in [cls.__name__ for cls in value.__class__.__mro__]:
            raise TypeError
        self._parent_chord = value
        # self._parent = value

    def _update_parent_note(self):
        if self.parent_note:
            self.parent_note._update_xml_pitch_or_rest()
            self.parent_note._update_xml_accidental()

    def _update_pitch_parameters(self):
        pitch = self.get_pitch_or_rest()
        if isinstance(pitch, XMLPitch):
            if not self.accidental.get_pitch_parameters()[1]:
                if pitch.xml_alter:
                    pitch.remove(pitch.xml_alter)
                pitch.xml_step, pitch.xml_octave = self.accidental.get_pitch_parameters()[0], \
                    self.accidental.get_pitch_parameters()[2]
            else:
                pitch.xml_step, pitch.xml_alter, pitch.xml_octave = self.accidental.get_pitch_parameters()
        else:
            raise TypeError

    def _update_pitch_or_rest(self):
        if self._pitch_or_rest is None:
            if self.value == 0:
                self._pitch_or_rest = XMLRest()
            else:
                self._pitch_or_rest = XMLPitch()
        else:
            if self.value == 0:
                if not isinstance(self._pitch_or_rest, XMLRest):
                    self._pitch_or_rest = XMLRest()
                    self._update_parent_note()
                else:
                    pass
            else:
                if not isinstance(self._pitch_or_rest, XMLPitch):
                    self._pitch_or_rest = XMLPitch()
                    self._update_parent_note()
                if self.accidental:
                    self.accidental._update_xml_object()
                    self._update_pitch_parameters()
        if self.up:
            self.up._update_xml_pitch_or_rest()

    # //public properties
    @property
    def accidental(self):
        return self._accidental

    @accidental.setter
    def accidental(self, value):
        if not value:
            value = Accidental()
        elif not isinstance(value, Accidental):
            raise TypeError(f'accidental.value must be of type Accidental not {type(value)}')
        self._accidental = value
        if value:
            self._accidental.parent_midi = self

    @property
    def notehead(self):
        return self._notehead

    @notehead.setter
    def notehead(self, val):
        if self.parent_note:
            raise AlreadyFinalizedError('Cannot change notehead after finalizing.')

        if val is None:
            self._notehead = None
        elif not isinstance(val, XMLNotehead):
            self._notehead = XMLNotehead(val)
        else:
            self._notehead = val

    @property
    def is_tied_to_next(self):
        if 'start' in self._ties:
            return True
        else:
            return False

    @property
    def is_tied_to_previous(self):
        if 'stop' in self._ties:
            return True
        else:
            return False

    @property
    def octave(self):
        return int(self.value / 12) - 1

    @property
    def parent_chord(self):
        return self._parent_chord

    @property
    def parent_note(self):
        return self._parent_note

    @parent_note.setter
    def parent_note(self, value):
        if value is not None and 'Note' not in [cls.__name__ for cls in value.__class__.__mro__]:
            raise TypeError
        self._parent_note = value
        self._parent = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        if not isinstance(v, float) and not isinstance(v, int):
            raise TypeError(f'Midi.value must be of type float or int not{type(v)}')
        if v != 0 and (v < 12 or v > 127):
            raise ValueError(
                f'Midi.value {v} can be zero for a rest or must be in a range between 12 and 127 inclusively')
        self._value = v
        self._update_pitch_or_rest()

    @property
    def name(self):
        """
        :returns a string like C#3 consisting of step, accidental sign (or value) and octave. Midi with value 0 returns 'rest' as its name.
        """
        if self.value == 0:
            return 'rest'

        pitch_step = self.accidental.get_pitch_parameters()[1]

        if not pitch_step:
            accidental = ''
        elif pitch_step == -1.5:
            accidental = 'b-'
        elif pitch_step == -1:
            accidental = 'b'
        elif pitch_step == -0.5:
            accidental = '-'
        elif pitch_step == 0.5:
            accidental = '+'
        elif pitch_step == 1:
            accidental = '#'
        elif pitch_step == 1.5:
            accidental = '#+'
        else:
            accidental = str(pitch_step)

        return f"{self.accidental.get_pitch_parameters()[0]}{accidental}{self.octave}"

    # //public methods
    def add_child(self, child: [Accidental]) -> Accidental:
        if not isinstance(child, Accidental):
            raise TypeError(f'child {child} must be of type Accidental.')
        if self.parent_chord and self.parent_chord._finalized is True:
            raise AlreadyFinalizedError(self, 'add_child')
        super().add_child(child)
        child._update()
        return child

    def add_tie(self, type_):
        if type_ not in ['start', 'stop']:
            raise ValueError
        self._ties.add(type_)
        if self.parent_note:
            self.parent_note._update_ties()

    def remove_tie(self, type_):
        removed = False
        try:
            self._ties.remove(type_)
            removed = True
        except KeyError:
            pass
        if removed and self.parent_note:
            self.parent_note._update_ties()

    def get_children(self) -> List[Accidental]:
        """
        :return: list of added children.
        :rtype: List[:obj:`~musicscore.accidental.Accidental`]
        """
        return super().get_children()

    def get_parent(self) -> 'Note':
        """
        :return: parent
        :rtype: :obj:`~musicscore.note.Note`
        """
        return super().get_parent()

    def get_pitch_or_rest(self) -> Union['XMLPitch', 'XMLRest']:
        """
        :return: XMLPitch with appropriate children.
        """
        return self._pitch_or_rest

    def get_staff_number(self):
        return self._staff_number

    def transpose(self, val: float) -> 'Midi':
        """
        Adds val to value and returns self
        """
        self.value += val
        return self

    def set_staff_number(self, val):
        self._staff_number = val

    # //operators
    def __lt__(self, other):  # For x < y
        return self.value < other.value

    def __le__(self, other):  # For x <= y
        return self.value <= other.value

    def __gt__(self, other):  # For x > y
        return self.value > other.value

    def __ge__(self, other):  # For x >= y
        return self.value >= other.value

    def __copy__(self):
        copied = self.__class__(value=self.value, accidental=self.accidental)
        copied._ties = self._ties
        return copied

    def __deepcopy__(self, memodict={}):
        copied_accidental = copy.copy(self.accidental)
        copied = self.__class__(value=self.value, accidental=copied_accidental)
        copied._ties = copy.copy(self._ties)
        return copied

    def _copy_for_split(self):
        copied_accidental = copy.copy(self.accidental)
        copied = self.__class__(value=self.value, accidental=copied_accidental)
        copied.notehead = self.notehead
        return copied


class MidiNote(Midi):
    """
    Parent class of shorthand midi names: C, D, E, F, G, A, B.
    Example C(4, '#') = C(4, 'sharp') = C(4, sharp)
    """
    _VALUE = 60

    def __init__(self, octave: object, accidental_sign: object = None, *args: object, **kwargs: object) -> object:
        self._accidental_sign = accidental_sign
        self._octave = octave
        super().__init__(value=self._get_value(), *args, **kwargs)
        self._set_accidental_mode()

    def _set_accidental_mode(self):
        self.accidental.mode = get_accidental_mode(self.value, self._accidental_sign)

    def _get_value(self):
        return self._VALUE + self._get_accidental_value() - (4 - self._octave) * 12

    def _get_accidental_value(self):
        permitted = ['double-flat', 'flat-flat', 'ff', 'bb', 'three-quarters-flat', 'flat', 'f', 'b', 'quarter-flat',
                     'quarter-sharp', 'sharp', 's', '#', 'three-quarters-sharp', 'double-sharp', 'sharp-sharp', 'ss',
                     '##', 'x']
        if self._accidental_sign in ['double-flat', 'flat-flat', 'ff', 'bb']:
            return -2
        elif self._accidental_sign in ['three-quarters-flat']:
            return -1.5
        elif self._accidental_sign in ['flat', 'f', 'b']:
            return -1
        elif self._accidental_sign in ['quarter-flat']:
            return -0.5
        elif self._accidental_sign in [None, 'natural']:
            return 0
        elif self._accidental_sign in ['quarter-sharp']:
            return 0.5
        elif self._accidental_sign in ['sharp', 's', '#']:
            return 1
        elif self._accidental_sign in ['three-quarters-sharp']:
            return 1.5
        elif self._accidental_sign in ['double-sharp', 'sharp-sharp', 'ss', '##', 'x']:
            return 2
        else:
            raise ValueError(f'accidental_sign value {self._accidental_sign} can be None or must be in {permitted}')

    def __repr__(self):
        return f"{self.name} at {id(self)}"

    def __copy__(self):
        copied = Midi(value=self.value, accidental=self.accidental)
        copied._ties = self._ties
        return copied

    def __deepcopy__(self, memodict={}):
        copied_accidental = copy.copy(self.accidental)
        copied = Midi(value=self.value, accidental=copied_accidental)
        copied._ties = copy.copy(self._ties)
        return copied

    def _copy_for_split(self):
        copied_accidental = copy.copy(self.accidental)
        copied = Midi(value=self.value, accidental=copied_accidental)
        copied.notehead = self.notehead
        return copied


class C(MidiNote):
    _VALUE = 60


class D(MidiNote):
    _VALUE = 62


class E(MidiNote):
    _VALUE = 64


class F(MidiNote):
    _VALUE = 65


class G(MidiNote):
    _VALUE = 67


class A(MidiNote):
    _VALUE = 69


class B(MidiNote):
    _VALUE = 71



"""
All <notation> elements defined in MusicXML. The lyrics show the notation assigned to each note.
"""
from pathlib import Path

from musicscore import Score, Chord, C, E, G
from musicscore.tests.util import IdTestCase
from musicscore.util import XML_DYNAMIC_CLASSES
from musicxml.xmlelement.xmlelement import XMLFermata, XMLArpeggiate, XMLNonArpeggiate, XMLAccidentalMark, XMLAccent, \
    XMLStrongAccent, XMLStaccato, XMLTenuto, XMLStaccatissimo, XMLSpiccato, XMLScoop, XMLPlop, XMLDoit, XMLFalloff, \
    XMLBreathMark, XMLCaesura, XMLStress, XMLUnstress, XMLTrillMark, XMLTurn, XMLDelayedTurn, XMLInvertedTurn, XMLShake, \
    XMLWavyLine, XMLDetachedLegato, XMLMordent, XMLInvertedMordent, XMLSchleifer, XMLTremolo, XMLUpBow, XMLDownBow, \
    XMLHarmonic, XMLNatural, XMLArtificial, XMLBasePitch, XMLTouchingPitch, XMLSoundingPitch, XMLOpenString, \
    XMLFingering, XMLThumbPosition, XMLPluck, XMLDoubleTongue, XMLTripleTongue, XMLStopped, XMLSnapPizzicato, XMLFret, \
    XMLString, XMLPullOff, XMLHammerOn, XMLBend, XMLBendAlter, XMLWithBar, XMLPreBend, XMLRelease, XMLTap, XMLHeel, \
    XMLToe, XMLFingernails, XMLF, XMLPpp, XMLSfp, XMLOtherDynamics


class TestLily321(IdTestCase):
    def test_lily_321_Notations(self):

        score = Score()
        part = score.add_part('p1')

        xml_notations = [
            XMLFermata(),
            XMLFermata('normal'),
            XMLFermata('angled'),
            XMLFermata('square'),
            XMLFermata(type='inverted'),
            XMLArpeggiate(),
            XMLNonArpeggiate(),
            XMLAccidentalMark('double-sharp'),
            XMLAccent(),
            XMLStrongAccent(),
            XMLStaccato(),
            XMLTenuto(),
            XMLDetachedLegato(),
            XMLStaccatissimo(),
            XMLSpiccato(),
            XMLScoop(),
            XMLPlop(),
            XMLDoit(),
            XMLFalloff(),
            XMLBreathMark(),
            XMLCaesura(),
            XMLStress(),
            XMLUnstress(),
            None,
            XMLTrillMark(),
            XMLTurn(),
            XMLDelayedTurn(),
            XMLInvertedTurn(),
            XMLShake(),
            [XMLWavyLine(type='start'), XMLWavyLine(type='continue'), XMLWavyLine(type='stop')],
            XMLMordent(),
            XMLInvertedMordent(),
            XMLSchleifer(),
            XMLTremolo(3),
            (XMLTurn(), XMLAccidentalMark('natural', placement='below')),
            (XMLTurn(), XMLAccidentalMark('sharp'), XMLAccidentalMark('three-quarters-flat')),
            XMLUpBow(),
            XMLDownBow(),
            XMLHarmonic(),
        ]

        harmonics = [XMLHarmonic() for _ in range(6)]
        harmonics[0].add_child(XMLNatural())
        harmonics[2].add_child(XMLArtificial())
        harmonics[3].add_child(XMLNatural())
        harmonics[3].add_child(XMLBasePitch())
        harmonics[4].add_child(XMLNatural())
        harmonics[4].add_child(XMLTouchingPitch())
        harmonics[5].add_child(XMLNatural())
        harmonics[5].add_child(XMLSoundingPitch())

        xml_notations += harmonics

        xml_notations += [
            XMLOpenString(),
            XMLThumbPosition()
        ]

        fingerings = [XMLFingering() for _ in range(7)]
        fingerings[1].value_ = '1'
        fingerings[2].value_ = '2'
        fingerings[3].value_ = '3'
        fingerings[4].value_ = '4'
        fingerings[5].value_ = '5'
        fingerings[5].value_ = 'something'
        fingerings.append((XMLFingering('2'), XMLFingering('3'), XMLFingering('5')))

        xml_notations += fingerings

        xml_notations += [
            XMLPluck(),
            XMLPluck('a'),
            XMLThumbPosition(),
            XMLDoubleTongue(),
            XMLTripleTongue(),
            XMLStopped(),
            XMLSnapPizzicato(),
            XMLFret(0),
            XMLFret(1),
            XMLString(1),
            XMLString(5),
            [XMLHammerOn(type='start'), XMLHammerOn(type='stop')],
            [XMLPullOff(type='start'), XMLPullOff(type='stop')]
        ]

        bends = [XMLBend() for _ in range(4)]
        bends[0].add_child(XMLBendAlter(3))
        bends[1].add_child(XMLBendAlter(3))
        bends[1].add_child(XMLWithBar())
        bends[2].add_child(XMLBendAlter(0.5))
        bends[2].add_child(XMLPreBend())
        bends[3].add_child(XMLBendAlter(3.5))
        bends[3].add_child(XMLRelease())

        xml_notations += bends

        xml_notations += [
            XMLTap(),
            XMLTap(hand='left'),
            XMLHeel(),
            XMLToe(),
            XMLFingernails(),
            None,
            None,
            None,
            XMLF(),
            XMLPpp(),
            XMLSfp(),
            XMLOtherDynamics('sfffz'),
            (XMLStrongAccent(), XMLStaccato()),
        ]

        for n in xml_notations:
            if n is None:
                ch = Chord(0, 1)
                part.add_chord(ch)
            elif isinstance(n, XMLNonArpeggiate):
                ch = Chord([C(5), E(5), G(5)], 1)
                ch.add_lyric('XMLNonArpeggiate not implemented')
                part.add_chord(ch)
            elif isinstance(n, XMLArpeggiate):
                ch = Chord([C(5), E(5), G(5)], 1)
                ch.add_x(n)
                part.add_chord(ch)
            elif isinstance(n, XMLAccidentalMark):
                ch = Chord(C(5), 1)
                ch.add_x(n, parent_type='notation')
                part.add_chord(ch)
            elif isinstance(n, list):
                for x in n:
                    ch = Chord(C(5), 1)
                    ch.add_x(x)
                    part.add_chord(ch)
            elif isinstance(n, tuple):
                ch = Chord(C(5), 1)
                for x in n:
                    if isinstance(x, XMLAccidentalMark):
                        ch.add_x(x, parent_type='ornament')
                    else:
                        ch.add_x(x)
                part.add_chord(ch)
            elif n.__class__ in XML_DYNAMIC_CLASSES:
                ch = Chord(C(5), 1)
                ch.add_x(n, parent_type='notation', placement='below')
                part.add_chord(ch)
            else:
                ch = Chord(C(5), 1)
                ch.add_x(n)
                part.add_chord(ch)

        objects = (XMLStaccato(), XMLAccent(), XMLTenuto())
        placements = ['above', 'below', 'below']
        ch = Chord(60, 1)
        for o, p in zip(objects, placements):
            ch.add_x(o, placement=p)
            print(o.placement)
        part.add_chord(ch)

        xml_path = Path(__file__).with_suffix('.xml')
        score.export_xml(xml_path)



from musicscore.exceptions import AlreadyFinalizedError, MusicTreeException


class FinalizeMixin:
    """
    FinalizeMixin is a mixin for finalizable classes: Score, Part, Measure, Staff, Voice, Beat and Chord
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._finalized = False

    def finalize(self) -> None:
        """
        finalize can only be called once.

        It calls finalize` method of all children.
        """
        if self._finalized:
            raise AlreadyFinalizedError(self)

        for child in self.get_children():
            if not child._finalized:
                child.finalize()

        self._finalized = True

    def to_string(self, *args, **kwargs):
        if not self._finalized:
            self.finalize()

        try:
            return super().to_string(*args, **kwargs)
        except AttributeError:
            raise MusicTreeException(
                f'{self.__class__} has no direct equivalent in MusicXML and cannot be converted to string.')



from typing import List

from musicscore.exceptions import TimeActualSignaturesNotValidError
from musicscore.quarterduration import QuarterDuration
from musicxml.xmlelement.xmlelement import XMLTime, XMLBeats, XMLBeatType
from musicscore.xmlwrapper import XMLWrapper

__all__ = ['Time', 'flatten_times']


def _convert_signatures_to_ints(signatures):
    output = []
    for i in range(int(len(signatures) / 2)):
        beat = signatures[i * 2]
        beat_type = signatures[i * 2 + 1]
        if isinstance(beat, str):
            l = beat.split('+')
            if len(l) == 1:
                output.append(int(beat))
                output.append(int(beat_type))
            else:
                for x in l:
                    output.append(int(x))
                    output.append(int(beat_type))
        else:
            output.append(beat)
            output.append(int(beat_type))
    if not output:
        raise TimeActualSignaturesNotValidError
    return output


def _get_quarter_durations_from_ints(signatures):
    output = 0
    for i in range(int(len(signatures) / 2)):
        beat = signatures[i * 2]
        beat_type = signatures[i * 2 + 1]
        output += 4 * beat / beat_type
    return output


class Time(XMLWrapper):
    _ATTRIBUTES = {'signatures', 'actual_signatures', 'parent_measure', 'show'}
    XMLClass = XMLTime

    def __init__(self, *signatures, show=True, **kwargs):
        super().__init__()
        self._xml_object = self.XMLClass(**kwargs)
        self.parent_measure = None

        self._signatures = None
        self._actual_signatures = None
        self._intern_actual_signatures = None
        self._show = None

        self.signatures = signatures
        self.show = show

    def _calculate_actual_signatures(self):
        map = {'2/8': [2, 8], '4/8': [2, 8, 2, 8], '5/8': [3, 8, 2, 8], '7/8': [4, 8, 3, 8]}
        signatures = _convert_signatures_to_ints(self.signatures)
        signatures = [signatures[i:i + 2] for i in range(0, len(signatures), 2)]
        self._intern_actual_signatures = []
        for signature in signatures:
            key = "/".join([str(x) for x in signature])
            if key in map:
                self._intern_actual_signatures.extend(map[key])
            elif signature[1] % 8 == 0 and signature[0] % 3 == 0:
                self._intern_actual_signatures.extend([3, signature[1]] * (signature[0] // 3))
            else:
                self._intern_actual_signatures.extend([1, signature[1]] * signature[0])
        return self._intern_actual_signatures

    def _reset_actual_signatures(self) -> None:
        """
        Resets actual signatures to None.
        """
        self._actual_signatures = None
        self._intern_actual_signatures = None

    def _update_signature_objects(self):
        signatures = [self.signatures[i:i + 2] for i in range(0, len(self.signatures), 2)]
        for beats, beat_type in zip(self._xml_object.find_children('XMLBeats'),
                                    self._xml_object.find_children('XMLBeatType')):
            if signatures:
                signature = signatures.pop(0)
                beats.value_ = str(signature[0])
                beat_type.value_ = str(signature[1])
            else:
                beats.up.remove(beats)
                beat_type.up.remove(beat_type)
        for beats, beat_type in signatures:
            self._xml_object.add_child(XMLBeats(str(beats)))
            self._xml_object.add_child(XMLBeatType(str(beat_type)))

    @property
    def actual_signatures(self) -> List[int]:
        """
        Set and gets actual signatures. If parent measure exists its beats inside voices will be updated.

        :type: Optional[List[int]]
        :return: A list of int representing actual time signatures. If not set manually, it is calculated internally. For example a 4/4 time
                 signature gets automatically [1, 4, 1, 4, 1, 4, 1, 4] as actual_signatures if not set otherwise.
        :rtype: List[int]
        """
        if self._actual_signatures is None:
            if self._intern_actual_signatures is None:
                self._calculate_actual_signatures()
            return self._intern_actual_signatures
        return self._actual_signatures

    @actual_signatures.setter
    def actual_signatures(self, val):
        if val is not None:
            val = _convert_signatures_to_ints(val)
        self._actual_signatures = val
        if self.parent_measure:
            self.parent_measure._update_voice_beats()

    @property
    def signatures(self) -> List[int]:
        """
        Set and gets signatures. If parent measure exists, beats inside its voices will be updated. If val is None, a 4/4 signature is set.

        :type: Optional[List[int]]
        :return: A list of int representing time signature. If not set manually, it is calculated internally. For example a 4/4 time
                 signature gets automatically [1, 4, 1, 4, 1, 4, 1, 4] as actual_signatures if not set otherwise.
        :rtype: List[int]
        """
        return self._signatures

    @signatures.setter
    def signatures(self, val):
        if not val:
            val = [4, 4]
        for v in val:
            if not isinstance(v, int) and not isinstance(v, str):
                raise TypeError
        self._signatures = val
        self._reset_actual_signatures()
        self._update_signature_objects()
        if self.parent_measure:
            self.parent_measure._update_voice_beats()

    @property
    def show(self) -> bool:
        """
        If time signature is shown or not.

        :type: bool
        :return: bool
        """
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError
        self._show = val

    @XMLWrapper.xml_object.getter
    def xml_object(self) -> XMLClass:
        return super().xml_object

    def get_beats_quarter_durations(self) -> List[QuarterDuration]:
        """
        :return: List of quarter durations according to :obj:`actual_signatures`
        """
        return [QuarterDuration(numerator, denominator) * 4 for numerator, denominator in
                [self.actual_signatures[i:i + 2] for i in range(0, len(self.actual_signatures), 2)]]

    def __copy__(self):
        cp = self.__class__(*self.signatures, show=self.show)
        cp._actual_signatures = self._actual_signatures
        return cp

    def __rmul__(self, other):
        return [self.__copy__() for _ in range(other)]


def flatten_times(times) -> List[Time]:
    """
    :param times: an expandable list of times or tuples representing times. For example [x * Time(3, 8)] return x time intances of Time(3, 8)
    :return: List[Time]

    >>> ts = [2 * Time(3, 8), (3, 4), 3 * [(1, 8)], Time(1, 8, 3, 4), Time(3, 4)]
    >>> [t.signatures for t in flatten_times(ts)]
    [(3, 8), (3, 8), (3, 4), (1, 8), (1, 8), (1, 8), (1, 8, 3, 4), (3, 4)]
    """
    output = []
    for time in times:
        if isinstance(time, Time):
            output.append(time)
        elif hasattr(time, '__iter__'):
            if {isinstance(t, int) for t in time} == {True}:
                output.append(Time(*time))
            else:
                for t in time:
                    if isinstance(t, Time):
                        output.append(t)
                    else:
                        output.append(Time(*t))
    return output



from pathlib import Path

from musicscore.chord import Chord
from musicscore.part import Part
from musicscore.score import Score
from musicscore.tests.util import IdTestCase, generate_all_quintuplets, generate_all_sextuplets, generate_all_triplets, \
    generate_all_septuplets


class TestTuplets1(IdTestCase):
    def test_generate_all_quintuplets(self):
        assert len(generate_all_quintuplets()) == 15
        for x in generate_all_quintuplets():
            assert sum(x) == 1

    def test_generate_all_sextuplets(self):
        assert len(generate_all_sextuplets()) == 27
        for x in generate_all_sextuplets():
            assert sum(x) == 1

    def test_tuplets_1(self):
        """
        Write all possible tuplet combinations up until SEXTUPLETS
        """

        s = Score()

        p = s.add_child(Part('P1', name='Music'))

        p.add_measure(time=(1, 4))

        tuplets = generate_all_triplets() + generate_all_quintuplets() + generate_all_sextuplets() + generate_all_septuplets()

        for tuplet_list in tuplets:
            for tuplet in tuplet_list:
                p.add_chord(Chord(60, tuplet))

        xml_path = Path(__file__).with_suffix('.xml')
        s.export_xml(xml_path)
